rcParam = {
    "compilable": True,
    "dump_masks_to_local": False,
}

try:  # pragma: no cover
    from numba import njit

    rcParam["compilable"] = True
except ImportError:  # pragma: no cover
    rcParam["compilable"] = False


def compileable(func):  # pragma: no cover
    """Decorate function to compile them using numba when available."""
    if rcParam["compilable"]:
        return njit(func)
    else:
        return func


def compileable_parallel(func):  # pragma: no cover
    """Decorate function to compile them (parallel) using numba when available."""
    if rcParam["compilable"]:
        return njit(func, parallel=True)
    else:
        return func



import numpy as np
import pytest
import xarray as xr

import seaduck as sd


def streamfunction(x, z):
    return np.cos(np.pi * x / 2) * np.cos(np.pi * z / 2)


@pytest.fixture
def overturning_cell():
    N = 100
    M = 50
    x = np.linspace(-1, 1, N + 1)
    y = np.linspace(-0.1, 0.1, 2)
    zl = np.linspace(0, -1000, M)
    zp1 = np.append(zl, -1001)
    xg, yg = np.meshgrid(x, y)
    xv = 0.5 * (xg[:, 1:] + xg[:, :-1])
    yv = 0.5 * (yg[:, 1:] + yg[:, :-1])
    0.5 * (xg[1:] + xg[:-1])
    0.5 * (yg[1:] + yg[:-1])

    xc = 0.5 * (xv[1:] + xv[:-1])
    yc = 0.5 * (yv[1:] + yv[:-1])

    tempx, tempz = np.meshgrid(x, zl)
    strmf = streamfunction(tempx, -tempz / 500 - 1).reshape(len(zl), 1, -1)
    z = 0.5 * (zp1[1:] + zp1[:-1])
    zl = zp1[:-1]
    drf = np.abs(np.diff(zp1))
    u = np.zeros((M, 1, N + 1), float)
    u[:-1] = np.diff(strmf, axis=0)
    w = np.diff(strmf, axis=-1)
    v = np.zeros((M, 2, N), float)
    stream = np.zeros((M, 2, N + 1))
    stream[:] = strmf
    ds = xr.Dataset(
        coords=dict(
            XC=(["Y", "X"], xc),
            YC=(["Y", "X"], yc),
            XG=(["Yp1", "Xp1"], xg),
            YG=(["Yp1", "Xp1"], yg),
            Zl=(["Zl"], zl),
            Z=(["Z"], z),
            drF=(["Z"], drf),
            rA=(["Y", "X"], np.ones_like(xc, float)),
        ),
        data_vars=dict(
            UVELMASS=(["Z", "Y", "Xp1"], u),
            VVELMASS=(["Z", "Yp1", "X"], v),
            WVELMASS=(["Zl", "Y", "X"], w),
            streamfunc=(["Zl", "Yp1", "Xp1"], stream),
        ),
    )
    return sd.OceData(ds)


@pytest.fixture
def gyre():
    M, N = 50, 50
    x = np.linspace(-1, 1, N + 1)
    y = np.linspace(-1, 1, M + 1)
    xg, yg = np.meshgrid(x, y)
    strmf = streamfunction(xg, yg)
    xv = 0.5 * (xg[:, 1:] + xg[:, :-1])
    yv = 0.5 * (yg[:, 1:] + yg[:, :-1])
    0.5 * (xg[1:] + xg[:-1])
    0.5 * (yg[1:] + yg[:-1])

    xc = 0.5 * (xv[1:] + xv[:-1])
    yc = 0.5 * (yv[1:] + yv[:-1])

    u = np.diff(strmf, axis=0)
    v = -np.diff(strmf, axis=1)
    ds = xr.Dataset(
        coords=dict(
            XC=(["Y", "X"], xc),
            YC=(["Y", "X"], yc),
            XG=(["Yp1", "Xp1"], xg),
            YG=(["Yp1", "Xp1"], yg),
            rA=(["Y", "X"], np.ones_like(xc, float)),
        ),
        data_vars=dict(
            UVELMASS=(["Y", "Xp1"], u),
            VVELMASS=(["Yp1", "X"], v),
            streamfunc=(["Yp1", "Xp1"], strmf),
        ),
    )
    return sd.OceData(ds)


@pytest.fixture
def gknw():
    kkk = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])
    return sd.KnW(kkk, vkernel="linear", tkernel="nearest")


@pytest.mark.filterwarnings("ignore:invalid value encountered")
@pytest.mark.filterwarnings("ignore:divide by zero")
def test_vertical_streamfunc_conservation(overturning_cell, gknw):
    N = 100
    x = np.random.uniform(-1, 1, N)
    z = np.random.uniform(-0.1, -1000, N)
    pt = sd.Particle(
        x=x,
        y=np.zeros_like(x),
        z=z,
        t=np.zeros_like(x),
        data=overturning_cell,
        transport=True,
    )
    with pytest.warns(UserWarning):
        before = pt.interpolate("streamfunc", gknw)
        steps = 40
        stops, ps = pt.to_list_of_time(
            normal_stops=np.linspace(0, -2 * steps * N, steps), update_stops=[]
        )
        after = pt.interpolate("streamfunc", gknw)
        assert np.allclose(after, before, atol=1e-7)


@pytest.mark.filterwarnings("ignore:invalid value encountered")
@pytest.mark.filterwarnings("ignore:divide by zero")
def test_horizontal_streamfunction_conservation(gyre, gknw):
    num_particle = 100
    np.random.seed(2023)
    x = np.random.random(num_particle) * 2 - 1
    y = np.random.random(num_particle) * 2 - 1
    pt = sd.Particle(
        x=x, y=y, z=None, t=np.zeros_like(x), data=gyre, wname=None, transport=True
    )
    gknw.vkernel = "nearest"
    with pytest.warns(UserWarning):
        before = pt.interpolate("streamfunc", gknw)
        steps = 15
        stops, ps = pt.to_list_of_time(
            normal_stops=np.linspace(0, 2 * steps * 50, steps), update_stops=[]
        )
        after = ps[-1].interpolate("streamfunc", gknw)
        assert np.allclose(after, before, atol=1e-7)



import copy

import numpy as np
import pytest
from scipy.interpolate import interp1d

import seaduck as sd
from seaduck import utils

uknw = sd.lagrangian.uknw
vknw = sd.lagrangian.vknw
wknw = sd.lagrangian.wknw

uuknw = copy.deepcopy(uknw)
vvknw = copy.deepcopy(vknw)
wwknw = copy.deepcopy(wknw)

wrongZknw = sd.KnW()
wrongZknw.vkernel = "something not correct"
wrongTknw = sd.KnW()
wrongTknw.tkernel = "something not correct"


@pytest.fixture
def prefetched():
    od = sd.OceData(utils.get_dataset("ecco"))
    return np.array(od["SALT"][slice(1, 2)])


# use float number to make Particle
@pytest.fixture
def ep():
    od = sd.OceData(utils.get_dataset("ecco"))
    return sd.Particle(x=-37.5, y=10.4586420059204, z=-9.0, t=698155200.0, data=od)


@pytest.mark.parametrize("od", ["aviso"], indirect=True)
def test_tz_not_None(od):
    ap = sd.Position()
    ap.from_latlon(x=-37.5, y=-60.4586420059204, z=-9.0, t=698155200.0, data=od)
    assert isinstance(ap.rel, sd.ocedata.RelCoord)


@pytest.mark.parametrize("y,t", [(10.4586420059204, None), (None, 698155200.0)])
@pytest.mark.parametrize("od", ["ecco"], indirect=True)
def test_xyt_is_None(od, y, t):
    ap = sd.Position()
    ap.from_latlon(x=-37.5, y=y, z=-9.0, t=t, data=od)
    if y is None:
        assert ap.rx is None
    if t is None:
        assert ap.rt is None


@pytest.mark.parametrize(
    "knw,required",
    [
        (uknw, "all"),
        (uknw, "Z"),
        (uknw, ["Z"]),
    ],
)
def test_fatten(ep, knw, required):
    fattened_inds = ep.fatten(knw, required=required)
    assert isinstance(fattened_inds[0], np.ndarray)
    assert len(fattened_inds[0].shape) > 1


@pytest.mark.parametrize("varname,knw", [("WVELMASS", uknw), ("UVELMASS", wknw)])
def test_interp_vertical(ep, varname, knw):
    res = ep.interpolate(varname, knw)
    assert isinstance(res, np.ndarray)
    assert len(res.shape) == 1


def test_interp_with_NoneZ(ep):
    ep.rz = None
    ep.rzl = None
    uuknw.ignore_mask = False
    wwknw.ignore_mask = False
    vvknw.ignore_mask = False

    res = ep.interpolate(["UVELMASS", "WVELMASS", "VVELMASS"], [uuknw, wwknw, vvknw])

    uuknw.ignore_mask = True
    wwknw.ignore_mask = True
    vvknw.ignore_mask = True

    assert isinstance(res, list)


def test_no_face_with_mask(ep):
    vvknw.ignore_mask = False
    uuknw.ignore_mask = False
    (
        output_format,
        main_keys,
        prefetch_dict,
        main_dict,
        hash_index,
        hash_mask,
        hash_read,
        hash_weight,
    ) = ep._register_interpolation_input(("UVELMASS", "VVELMASS"), (uuknw, vvknw))
    index_lookup = ep._fatten_required_index_and_register(hash_index, main_dict)
    transform_lookup = ep._transform_vector_and_register(
        index_lookup, hash_index, main_dict
    )
    neo_transform_lookup = {}
    for key in transform_lookup.keys():
        neo_transform_lookup[key] = None
    mask_dict = ep._mask_value_and_register(
        index_lookup, neo_transform_lookup, hash_mask, hash_index, main_dict
    )
    assert bool(mask_dict)


def test_dict_input(ep):
    registered = ep._register_interpolation_input(
        "SALT",
        {"SALT": uknw},
        prefetched={"SALT": prefetched},
        prefetch_prefix={"SALT": (1, 0, 0, 0, 0)},
    )
    assert len(registered) == 8


@pytest.mark.parametrize(
    "varname,knw,num_prefetched",
    [
        ("SALT", uknw, "one"),
        (("SALT", "SALT"), (uknw, uknw), "two"),
    ],
)
def test_diff_prefetched(ep, prefetched, varname, knw, num_prefetched):
    prefetch = prefetched
    if num_prefetched == "two":
        prefetch = (prefetch, prefetch)
    registered = ep._register_interpolation_input(
        varname, knw, prefetched=prefetch, prefetch_prefix={"SALT": (1, 0, 0, 0, 0)}
    )
    assert len(registered) == 8


@pytest.mark.parametrize(
    "data,x,knw",
    [
        (None, -37.5, uknw),
        (prefetched, -37.5, uknw),
    ],
)
@pytest.mark.parametrize("od", ["ecco"], indirect=True)
def test_wrong_datatype_error(od, data, x, knw):
    with pytest.raises(ValueError):
        the_p = sd.Particle(
            x=x,
            y=np.ones(2) * 10.4586420059204,
            z=np.ones(2) * (-9.0),
            t=np.ones(2) * 698155200.0,
            data=data,
        )
        the_p.fatten(knw)


@pytest.mark.parametrize(
    "x,knw",
    [
        (np.ones((2, 2)), uknw),
        (np.ones(12), uknw),
        (np.array([-37.5, -37.4]), wrongZknw),
        (np.array([-37.5, -37.4]), wrongTknw),
    ],
)
@pytest.mark.parametrize("od", ["ecco"], indirect=True)
def test_init_valueerror(od, x, knw):
    with pytest.raises(ValueError):
        the_p = sd.Particle(
            x=x,
            y=np.ones(2) * 10.4586420059204,
            z=np.ones(2) * (-9.0),
            t=np.ones(2) * 698155200.0,
            data=od,
        )
        the_p.fatten(knw)


@pytest.mark.parametrize(
    "varName,knw,prefetch,prefetch_prefix",
    [
        (1, uknw, None, None),
        ("SALT", [uknw, vknw], None, None),
        (("SALT", "SALT"), (uknw, vknw, wknw), None, None),
        ("SALT", {"s": uknw}, None, None),
        ("SALT", 1, None, None),
        ("SALT", uknw, [1, 1], None),
        ("SALT", uknw, None, [1, 1]),
        ("SALT", uknw, 1, None),
        ("SALT", uknw, None, 1),
    ],
)
def test_interp_register_error(ep, varName, knw, prefetch, prefetch_prefix):
    with pytest.raises(ValueError):
        ep._register_interpolation_input(
            varName, knw, prefetched=prefetch, prefetch_prefix=prefetch_prefix
        )


def test_fatten_none(ep):
    ep.it = None
    ep.izl = None
    ep.iz = None

    assert ep._fatten_v(wknw) is None
    assert ep._fatten_vl(wknw) is None
    assert ep._fatten_t(wknw) is None


def test_partial_flatten():
    thing = np.array((3, 4, 5))
    ind = (thing, thing)
    flattened = sd.eulerian._partial_flatten(ind)
    assert flattened[0].shape == (3, 1)


@pytest.mark.parametrize("od", ["ecco"], indirect=True)
def test_wvel_quant_deepest(od):
    ind = (11, 75, 73)
    face, iy, ix = ind

    np.random.seed(20230625)
    z = np.random.uniform(od.Zl[-1], 0, 5000)
    x = od.XC[ind] * np.ones_like(z)
    y = od.YC[ind] * np.ones_like(z)

    vert_p = sd.Position().from_latlon(x=x, y=y, z=z, data=od)
    assert vert_p.ix[0] == ix, "horizontal index does not match"
    assert vert_p.iy[0] == iy, "horizontal index does not match"
    seaduck_ans = vert_p.interpolate("WVELMASS1", sd.lagrangian.wknw)

    wvel = interp1d(od.Zl, od._ds.WVELMASS1[:, face, iy, ix])
    scipy_ans = wvel(z)

    assert np.allclose(scipy_ans, seaduck_ans)


@pytest.mark.parametrize("od", ["ecco"], indirect=True)
@pytest.mark.parametrize("seed", list(range(5)))
def test_wvel_quant_random_place(od, seed):
    np.random.seed(seed)
    z = np.random.uniform(od.Zl[-1], 0, 50)
    x = np.random.uniform(-180, 180, 1) * np.ones_like(z)
    y = np.random.uniform(-80, 90, 1) * np.ones_like(z)

    vert_p = sd.Position().from_latlon(x=x, y=y, z=z, data=od)
    seaduck_ans = vert_p.interpolate("WVELMASS1", sd.lagrangian.wknw)

    face = vert_p.face[0]
    iy = vert_p.iy[0]
    ix = vert_p.ix[0]

    wvel = interp1d(od.Zl, od._ds.WVELMASS1[:, face, iy, ix])
    scipy_ans = wvel(z)

    assert np.allclose(scipy_ans, seaduck_ans)


@pytest.mark.parametrize("od", ["ecco"], indirect=True)
def test_dw_quant_deepest(od):
    ind = (11, 75, 73)
    face, iy, ix = ind

    np.random.seed(20230625)
    z = np.random.uniform(od.Zl[-1], 0, 5000)
    x = od.XC[ind] * np.ones_like(z)
    y = od.YC[ind] * np.ones_like(z)

    vert_p = sd.Position().from_latlon(x=x, y=y, z=z, data=od)
    assert vert_p.ix[0] == ix, "horizontal index does not match"
    assert vert_p.iy[0] == iy, "horizontal index does not match"
    seaduck_ans = vert_p.interpolate("WVELMASS1", sd.lagrangian.dwknw)

    # dw is a stepwise function.
    small_offset = 1e-12
    dw = -np.diff(np.array(od._ds.WVELMASS1[:, face, iy, ix]))
    zinterp = [0.0]
    dwinterp = [dw[0]]
    for i, zl in enumerate(od.Zl[1:-1]):
        zinterp.append(float(zl) + small_offset)
        dwinterp.append(dw[i])
        zinterp.append(float(zl))
        dwinterp.append(dw[i + 1])
    zinterp.append(float(od.Zl[-1]))
    dwinterp.append(dw[-1])
    dwvel = interp1d(zinterp, dwinterp)
    scipy_ans = dwvel(z)

    assert np.allclose(scipy_ans, seaduck_ans)


@pytest.mark.parametrize("od", ["ecco"], indirect=True)
@pytest.mark.parametrize("seed", list(range(7, 12)))
def test_dw_quant_random(od, seed):
    np.random.seed(seed)
    z = np.random.uniform(od.Zl[-1], 0, 50)
    x = np.random.uniform(-180, 180, 1) * np.ones_like(z)
    y = np.random.uniform(-80, 90, 1) * np.ones_like(z)

    vert_p = sd.Position().from_latlon(x=x, y=y, z=z, data=od)
    seaduck_ans = vert_p.interpolate("WVELMASS1", sd.lagrangian.dwknw)

    face = vert_p.face[0]
    iy = vert_p.iy[0]
    ix = vert_p.ix[0]

    # dw is a stepwise function.
    small_offset = 1e-12
    dw = -np.diff(np.array(od._ds.WVELMASS1[:, face, iy, ix]))
    zinterp = [0.0]
    dwinterp = [dw[0]]
    for i, zl in enumerate(od.Zl[1:-1]):
        zinterp.append(float(zl) + small_offset)
        dwinterp.append(dw[i])
        zinterp.append(float(zl))
        dwinterp.append(dw[i + 1])
    zinterp.append(float(od.Zl[-1]))
    dwinterp.append(dw[-1])
    dwvel = interp1d(zinterp, dwinterp)
    scipy_ans = dwvel(z)

    assert np.allclose(scipy_ans, seaduck_ans)



import copy
import logging
import warnings

import numpy as np
import xarray as xr

from seaduck.ocedata import RelCoord
from seaduck.smart_read import smart_read


def mask_u_node(maskC, tp):
    """Mask out U-points.

    for MITgcm indexing, U is defined on the left of the cell,
    When the C grid is dry, U are either:
    a. dry;
    b. on the interface, where the cell to the left is wet.
    if b is the case, we need to unmask the udata, because it makes some physical sense.

    Parameters
    ----------
    maskC: numpy.ndarray
        numpy array with the same shape as the model spacial coordinates.
        1 for wet cells (center points), 0 for dry ones.
    tp: Topology object
        The Topology object for the dataset.

    Returns
    -------
    maskU: numpy.ndarray
        numpy array with the same shape as the model spacial coordinates.
        1 for wet U-walls (including interface between wet and dry), 0 for dry ones.
    """
    maskU = copy.deepcopy(maskC)
    indexes = np.array(np.where(maskC == 0)).T
    # find out which points are masked will make the search faster.
    new_ind = tp.ind_tend_vec(indexes.T[1:], np.ones_like(indexes.T[0], int) * 2)
    new_ind = np.vstack([indexes.T[0], new_ind])
    switch = indexes[np.where(maskC[tuple(new_ind)])]
    maskU[tuple(switch.T)] = 1

    return maskU


def mask_v_node(maskC, tp):
    """Mask out v-points.

    for MITgcm indexing, V is defined on the "south" side of the cell,
    When the C grid is dry, V are either:
    a. dry;
    b. on the interface, where the cell to the downside is wet.
    if b is the case, we need to unmask the vdata.

    Parameters
    ----------
    maskC: numpy.ndarray
        numpy array with the same shape as the model spacial coordinates.
        1 for wet cells (center points), 0 for dry ones.
    tp: Topology object
        The Topology object for the dataset.

    Returns
    -------
    + maskV: numpy.ndarray
        numpy array with the same shape as the model spacial coordinates.
        1 for wet W-walls (including interface between wet and dry), 0 for dry ones.
    """
    maskV = copy.deepcopy(maskC)
    indexes = np.array(np.where(maskC == 0)).T
    # find out which points are masked will make the search faster.
    new_ind = tp.ind_tend_vec(indexes.T[1:], np.ones_like(indexes.T[0], int) * 1)
    new_ind = np.vstack([indexes.T[0], new_ind])
    switch = indexes[np.where(maskC[tuple(new_ind)])]
    maskV[tuple(switch.T)] = 1
    return maskV


def mask_w_node(maskC, tp=None):
    # this one does not need tp object
    # if you pass something into it by mistake, it will be ignored.
    """Mask out W-points.

    for MITgcm indexing, W is defined on the top of the cell,
    When the C grid is dry, W are either:
    a. dry;
    b. on the interface, where the cell above is wet.
    if b is the case, we need to unmask the wdata.

    Parameters
    ----------
    maskC: numpy.ndarray
        numpy array with the same shape as the model spacial coordinates.
        1 for wet cells (center points), 0 for dry ones.
    tp: Topology object
        The Topology object for the dataset.

    Returns
    -------
    + maskWvel: numpy.ndarray
        numpy array with the same shape as the model spacial coordinates.
        1 for wet W-walls (including interface between wet and dry), 0 for dry ones.
    """
    maskW = np.zeros_like(maskC)
    maskW[1:] = maskC[:-1]
    maskW = np.logical_or(maskW, maskC).astype(int)
    return maskW


def get_mask_arrays(od):
    """Mask all staggered valocity points.

    A wrapper around mask_u_node, mask_v_node, mask_w_node.
    If there is no maskC in the dataset, just return nothing is masked.

    Parameters
    ----------
    od: OceData object
        The dataset to compute masks on.
    tp: Topology object
        The Topology of the datset

    Returns
    -------
    maskC,maskU,maskV,maskW: numpy.ndarray
        masks at center points, U-walls, V-walls, W-walls respectively.
    """
    tp = od.tp
    keys = od._ds.keys()
    if "maskC" not in keys:
        warnings.warn("no maskC in the dataset, assuming nothing is masked.")
        logging.warning("no maskC in the dataset, assuming nothing is masked.")
        maskC = np.ones_like(od._ds.XC + od._ds.Z)
        # it is inappropriate to fill in the dataset,
        # expecially given that there is no performance boost.
        return maskC, maskC, maskC, maskC
    maskC = np.array(od._ds["maskC"])
    if "Z" not in od._ds["maskC"].dims:
        raise NotImplementedError(
            "2D land mask is not yet supported,"
            "you could potentially get around by adding a psuedo dimension"
            "or you could set knw.ignore_mask = True"
        )
    if "maskU" not in keys:
        logging.info("creating maskU,this is going to be very slow!")
        maskU = mask_u_node(maskC, tp)
        od._ds["maskU"] = od._ds["Z"] + od._ds["XG"]
        od._ds["maskU"].values = maskU
    else:
        maskU = np.array(od._ds["maskU"])
    if "maskV" not in keys:
        logging.info("creating maskV,this is going to be very slow!")
        maskV = mask_v_node(maskC, tp)
        od._ds["maskV"] = od._ds["Z"] + od._ds["YG"]
        od._ds["maskV"].values = maskV
    else:
        maskV = np.array(od._ds["maskV"])
    if "maskWvel" not in keys:
        # there is a maskW with W meaning West in ECCO
        logging.info("creating maskW,this is going to be somewhat slow")
        maskW = mask_w_node(maskC)
        od._ds["maskWvel"] = od._ds["Z"] + od._ds["YC"]
        od._ds["maskWvel"].values = maskW
    else:
        maskW = np.array(od._ds["maskWvel"])
    return maskC, maskU, maskV, maskW


def get_masked(od, ind, cuvwg="C"):
    """Return whether points are masked.

    Return whether the indexes of intersts are masked or not.

    Parameters
    ----------
    od: OceData object
        Dataset to find mask values from.
    ind: tuple of numpy.ndarray
        Indexes of grid points.
    cuvwg: str
        Whether the indexes is for points at center points or on the walls.
        Options are: ['C','U','V','Wvel'].
    """
    if cuvwg not in ["C", "U", "V", "Wvel"]:
        raise NotImplementedError(
            "cuvwg(the kind of grid point) for mask not supported"
        )
    keys = od._ds.keys()
    if "maskC" not in keys:
        warnings.warn("no maskC in the dataset, assuming nothing is masked.")
        return np.ones_like(ind[0])
    elif cuvwg == "C":
        return smart_read(od._ds.maskC, ind)

    name = "mask" + cuvwg
    tp = od.tp
    maskC = np.array(od._ds["maskC"])
    func_dic = {"U": mask_u_node, "V": mask_v_node, "Wvel": mask_w_node}
    rename_dic = {
        "U": lambda x: x if x != "X" else "Xp1",
        "V": lambda x: x if x != "Y" else "Xp1",
        "Wvel": lambda x: x if x != "Z" else "Zl",
    }
    if name not in keys:
        small_mask = func_dic[cuvwg](maskC, tp)
        dims = tuple(map(rename_dic[cuvwg], od._ds.maskC.dims))
        sizes = tuple(len(od._ds[dim]) for dim in dims)
        mask = np.zeros(sizes)
        # indexing sensitive
        old_size = small_mask.shape
        slices = tuple(slice(0, i) for i in old_size)
        mask[slices] = small_mask
        od._ds[name] = xr.DataArray(mask, dims=dims)
        return mask[ind]
    else:
        return smart_read(od._ds[name], ind)


def which_not_stuck(p):
    """Investigate which points are in land mask."""
    ind = []
    if p.izl_lin is not None:
        ind.append(p.izl_lin - 1)
    if p.face is not None:
        ind.append(p.face)
    ind += [p.iy, p.ix]
    ind = tuple(ind)
    return get_masked(p.ocedata, ind).astype(bool)


def abandon_stuck(p):
    """Abandon those stucked in mud."""
    which = which_not_stuck(p)
    vardict = vars(p)
    keys = vardict.keys()
    for i in keys:
        item = vardict[i]
        if isinstance(item, np.ndarray):
            if len(item.shape) == 1:
                setattr(p, i, item[which])
                p.N = len(getattr(p, i))
            else:
                setattr(p, i, item)
        elif isinstance(item, RelCoord):
            setattr(p, i, item.subset(which))
        else:
            setattr(p, i, item)
    return p



import pytest

import seaduck as sd
from seaduck import utils


@pytest.fixture(scope="session")
def ds(request):
    return utils.get_dataset(request.param)


@pytest.fixture(scope="session")
def od(request):
    return sd.OceData(utils.get_dataset(request.param))


@pytest.fixture(scope="session")
def tp(request):
    return sd.Topology(utils.get_dataset(request.param))



import logging

import numpy as np

try:  # pragma: no cover
    import pandas as pd
except ImportError:  # pragma: no cover
    pass
import xarray as xr

from seaduck.topology import Topology
from seaduck.utils import (
    _general_len,
    create_tree,
    find_rel_h_naive,
    find_rel_h_oceanparcel,
    find_rel_h_rectilinear,
    find_rel_nearest,
    find_rel_time,
    find_rel_z,
    missing_cs_sn,
)

NO_ALIAS = {
    "dXC": "dxC",
    "dYC": "dyC",
    "dZ": "drC",
    "dXG": "dxG",
    "dYG": "dyG",
    "dZl": "drF",
}


class RelCoord(dict):
    """
    NamedTuple that also has update method.

    This class is used to store the relative coordinates.
    Attributes starts with "i" are indexes of the nearest grid point.
    Attributes starts with "b" are value (time/dep/lat/lon) of the nearest grid point.
    Attributes starts with "d" are distance between the nearest grid point and its
    neighboring point in meters or seconds.
    Attributes starts with "r" are the distance from the point of interest to the nearest
    non-dimensionalized by the "d" variable.
    "cs", "sn" are the cosine and sine of the grid orientation relative to meridian.
    "face" is the face/tile the point is on, if the dataset has such a dimension.

    All of those attributes should be None or 1D numpy array.

    Methods
    -------
    update(other)
        Inheritated from dictionary.
    """

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError as exc:
            raise AttributeError(
                f"'RelCoord' object has no attribute '{attr}'"
            ) from exc

    def __setattr__(self, attr, value):
        self[attr] = value

    def subset(self, which):
        """Create a subset of all the non-None items."""
        new = RelCoord()
        for var in self.keys():
            if self[var] is not None:
                new[var] = self[var][which]
            else:
                new[var] = None
        return new

    def update_from_subset(self, sub, which):
        for var in sub.keys():
            if var in self.keys():
                if self[var] is not None:
                    self[var][which] = sub[var]
            else:
                raise KeyError(
                    f"Key {var} is in subset, " "but not in the one to update."
                )

    @classmethod
    def create_class(cls, class_name, fields):
        """Create a subclass with predetermined keys."""

        class NewClass(cls):
            __slots__ = ()
            _fields = fields

            def __init__(self, *args):
                if len(args) > len(fields):
                    raise TypeError(
                        f"{class_name} takes {len(fields)} positional arguments"
                        f" but {len(args)} were given"
                    )
                for name, value in zip(fields, args):
                    setattr(self, name, value)

            @classmethod
            def _make(cls, iterable):
                """Make a instance of this class similar to that of collections.namedtuple."""
                return cls(*iterable)

            def __repr__(self):
                field_values = ", ".join(
                    f"{name}={getattr(self, name)!r}" for name in self._fields
                )
                return f"{class_name}({field_values})"

        NewClass.__name__ = class_name
        return NewClass


HRel = RelCoord.create_class(
    "HRel", ["face", "iy", "ix", "rx", "ry", "cs", "sn", "dx", "dy", "bx", "by"]
)
HRel.__doc__ = "Wrap around the horizontal rel-coords. See also RelCoord."
VRel = RelCoord.create_class("VRel", ["iz", "rz", "dz", "bz"])
VRel.__doc__ = (
    "Wrap around the vertical centered nearest rel-coords. See also RelCoord."
)
VLinRel = RelCoord.create_class("VLRel", ["iz_lin", "rz_lin", "dz_lin", "bz_lin"])
VLinRel.__doc__ = (
    "Wrap around the vertical centered linear rel-coords. See also RelCoord."
)
VlRel = RelCoord.create_class("VlRel", ["izl", "rzl", "dzl", "bzl"])
VlRel.__doc__ = (
    "Wrap around the vertical staggered nearest rel-coords. See also RelCoord."
)
VlLinRel = RelCoord.create_class(
    "VlLinRel", ["izl_lin", "rzl_lin", "dzl_lin", "bzl_lin"]
)
VlLinRel.__doc__ = (
    "Wrap around the vertical staggered linear rel-coords. See also RelCoord."
)
TRel = RelCoord.create_class("TRel", ["it", "rt", "dt", "bt"])
TRel.__doc__ = "Wrap around the temporal nearest rel-coords."
TLinRel = RelCoord.create_class("TLinRel", ["it_lin", "rt_lin", "dt_lin", "bt_lin"])
TRel.__doc__ = "Wrap around the temporal linear rel-coords. See also RelCoord."


class OceData:
    """Ocean dataset.

    Class that enables variable aliasing, topology extraction, and 4-D translation
    between latitude-longitude-depth-time grid and relative description.

    Parameters
    ----------
    data: xarray.Dataset
        The dataset to extract grid information, create cKD tree, and Topology object on.
    alias: dict, None, or 'auto'
        1. dict: Map the variable used by this package (key) to
           that used by the dataset (value).
        2. None (default): Do not apply alias.
        3. 'auto' (Not implemented): Automatically generate a list for the alias.
    """

    def __init__(self, data, alias=None, memory_limit=1e7):
        self._ds = data.transpose(
            "time",
            "time_midp",
            "time_outer",
            "Z",
            "Zl",
            "Zp1",
            "face",
            "Y",
            "Yp1",
            "X",
            "Xp1",
            ...,
            missing_dims="ignore",
        )
        self.tp = Topology(self._ds)
        if alias is None:
            self.alias = NO_ALIAS
        elif alias == "auto":
            raise NotImplementedError("auto alias not yet implemented")
        elif isinstance(alias, dict):
            self.alias = alias

        try:
            self.too_large = self._ds["XC"].nbytes > memory_limit
        except KeyError:
            self.too_large = False
        readiness, missing = self.check_readiness()
        self.readiness = readiness
        if readiness:
            self._grid2array()
        else:
            raise ValueError(
                f"use add_missing_variables or set_alias to create {missing},"
                "then call OceData.grid2array."
            )
        if self.readiness["Zl"]:
            # make a more consistent vector definition
            with_zl = [i for i in self._ds.data_vars if "Zl" in self._ds[i].dims]
            if len(with_zl) > 0:
                without_zl = [i for i in self._ds.data_vars if i not in with_zl]
                bottom_buffer = xr.zeros_like(self._ds[with_zl].isel(Zl=slice(1)))
                bottom_buffer["Zl"] = [self.Zl[-1]]
                neods = xr.merge(
                    [
                        xr.concat([self._ds[with_zl], bottom_buffer], dim="Zl"),
                        self._ds[without_zl],
                    ]
                )
                self._ds = neods

    def __setitem__(self, key, item):
        if isinstance(item, xr.DataArray):
            if key in self.alias.keys():
                self._ds[self.alias[key]] = item
            else:
                self._ds[key] = item
        else:
            object.__setattr__(self, key, item)

    def __getitem__(self, key):
        varsdict = vars(self)
        if key in varsdict.keys():
            return object.__getattribute__(self, key)
        else:
            if key in self.alias.keys():
                return self._ds[self.alias[key]]
            else:
                return self._ds[key]

    def check_readiness(self):
        """Return readiness.

        Function that checks what kind of interpolation is supported.

        Returns
        -------
        readiness: dict
            1. h: The scheme of horizontal interpolation to be used, including
               oceanparcel, local_cartesian, and 'rectilinear'.
            2. Z: Whether the dataset has a vertical dimension at the center points.
            3. Zl: Whether the dataset has a vertical dimension at
               staggered points (vertical velocity).
            4. time: Whether the dataset has a temporal dimension.
        missing: list
            The missing variable names.
        """
        # TODO: make the check more detailed
        varnames = list(self._ds.data_vars) + list(self._ds.coords)
        readiness = {}
        missing = []
        if all(i in varnames for i in ["XC", "YC", "XG", "YG"]):
            readiness["h"] = "oceanparcel"
            # could be a curvilinear grid that can use oceanparcel style
        elif all(i in varnames for i in ["XC", "YC", "dxG", "dyG", "CS", "SN"]):
            readiness["h"] = "local_cartesian"
            # could be a curvilinear grid that can use local cartesian style
        elif all(i in varnames for i in ["lon", "lat"]):
            ratio = 6371e3 * np.pi / 180
            self.dlon = np.gradient(self["lon"]) * ratio
            self.dlat = np.gradient(self["lat"]) * ratio
            readiness["h"] = "rectilinear"
            # corresponding to a rectilinear dataset
        else:
            readiness["h"] = False
            # readiness['overall'] = False
            missing.append(
                """
            For the most basic use case,
            {XC,YC,XG,YG} or {XC,YC,dxG,dyG,CS,SN} for curvilinear grid;
            or {lon, lat} for rectilinear grid.
            """
            )
            return False, missing
        for _ in ["time", "Z", "Zl"]:
            readiness[_] = (_ in varnames) and (_general_len(self[_]) > 1)

        return readiness, missing

    def _add_missing_grid(self):
        # TODO:
        pass

    def show_alias(self):
        """Print out the renaming in a nice pd.DataFrame format.

        Returns
        -------
        pretty_alias: pandas.DataFrame
            Carry the same information as self.alias,
            look a little nicer.
        """
        return pd.DataFrame.from_dict(
            self.alias, orient="index", columns=["original name"]
        )

    def _add_missing_cs_sn(self):
        try:
            assert self["SN"] is not None
            assert self["CS"] is not None
        except (AttributeError, AssertionError):
            cs, sn = missing_cs_sn(self._ds)
            self["CS"] = cs
            self["SN"] = sn

    def _add_missing_vol(self, as_numpy=False):
        if self.readiness["Zl"]:
            vol = self._ds["drF"] * self._ds["rA"]
            if "HFacC" in self._ds.data_vars:
                vol *= self._ds["HFacC"]
        else:
            vol = self._ds["rA"]
        vol = vol.fillna(0)
        if as_numpy:
            self["Vol"] = np.array(vol)
        else:
            self["Vol"] = vol

    def _hgrid2array(self):
        """Extract the horizontal grid data into numpy arrays.

        This is done based on the readiness['h'] of the OceData object.
        """
        if self.too_large:  # pragma: no cover
            logging.warning(
                "Loading grid into memory, it's a large dataset please be patient"
            )

        if self.readiness["h"] == "oceanparcel":
            for var in ["XC", "YC", "XG", "YG"]:
                self[var] = np.array(self[var], dtype="float32")
            for var in ["rA", "CS", "SN"]:
                try:
                    self[var] = np.array(self[var], dtype="float32")
                except KeyError:
                    logging.info("no %s in dataset, skip", var)
                    self[var] = None
            try:
                self.dX = np.array(self["dXG"], dtype="float32")
                self.dY = np.array(self["dYG"], dtype="float32")
            except KeyError:
                self.dX = None
                self.dY = None
            if self.too_large:  # pragma: no cover
                logging.info("numpy arrays of grid loaded into memory")
            self.tree = create_tree(self.XC, self.YC)
            if self.too_large:  # pragma: no cover
                logging.info("cKD created")

        if self.readiness["h"] == "local_cartesian":
            for var in ["XC", "YC", "CS", "SN"]:
                self[var] = np.array(self[var], dtype="float32")
            self.dX = np.array(self["dXG"], dtype="float32")
            self.dY = np.array(self["dYG"], dtype="float32")

            if not self.too_large:  # pragma: no cover
                for var in ["XG", "YG", "dXC", "dYC", "rA"]:
                    try:
                        self[var] = np.array(self[var], dtype="float32")
                    except KeyError:
                        logging.info("no %s in dataset, skip", var)
                        self[var] = None
            if self.too_large:  # pragma: no cover
                logging.info("numpy arrays of grid loaded into memory")
            self.tree = create_tree(self.XC, self.YC)
            if self.too_large:  # pragma: no cover
                logging.info("cKD created")

        if self.readiness["h"] == "rectilinear":
            self.lon = np.array(self["lon"], dtype="float32")
            self.lat = np.array(self["lat"], dtype="float32")

    def _vgrid2array(self):
        """Extract the vertical center point grid data into numpy arrays."""
        self.Z = np.array(self["Z"], dtype="float32")
        try:
            self.dZ = np.array(self["dZ"], dtype="float32")
        except KeyError:
            self.dZ = np.diff(self.Z)
            self.dZ = np.append(self.dZ, self.dZ[-1]).astype("float32")

    def _vlgrid2array(self):
        """Extract the vertical staggered point grid data into numpy arrays."""
        if "Zp1" in self._ds.variables:
            self.Zl = np.array(self["Zp1"], dtype="float32")
        else:
            self.Zl = np.zeros(len(self["Zl"]) + 1)
            self.Zl[:-1] = np.array(self._ds["Zl"])
            self.Zl[-1] = 2 * self.Zl[-2] - self.Zl[-3]
            self.Zl = self.Zl.astype("float32")
        try:
            self.dZl = np.array(self["dZl"], dtype="float32")
        except KeyError:
            self.dZl = np.diff(self.Zl).astype("float32")

        # special treatment for dZl
        # self.dZl = np.roll(self.dZl,1)
        # self.dZl[0] = 1e-10

    def _tgrid2array(self):
        """Extract the temporal grid data into numpy arrays."""
        self.t_base = 0
        self.ts = np.array(self["time"])
        if self["time"].dtype != "float":
            self.ts = (self.ts).astype(float) / 1e9
        try:
            self.time_midp = np.array(self["time_midp"])
            self.time_midp = (self.time_midp).astype(float) / 1e9
        except KeyError:
            self.time_midp = (self.ts[1:] + self.ts[:-1]) / 2

    def _grid2array(self):
        """Assemble all the extraction methods."""
        if self.readiness["h"]:
            self._hgrid2array()
        if self.readiness["Z"]:
            self._vgrid2array()
        if self.readiness["Zl"]:
            self._vlgrid2array()
        if self.readiness["time"]:
            self._tgrid2array()

    def _find_rel_h(self, x, y):
        """Find the horizontal rel-coordinate.

        Find the horizontal rel-coordinate of the given 4-D position based on readiness['h'].

        Parameters
        ----------
        x, y: np.ndarray
            1D array of longitude and latitude.

        Returns
        -------
        hrel: seaduck.ocedata.HRel object
            A dictionary that defines the horizontal rel-coords
        """
        if self.readiness["h"] == "oceanparcel":
            h_rel_tuple = find_rel_h_oceanparcel(
                x,
                y,
                self.XC,
                self.YC,
                self.dX,
                self.dY,
                self.CS,
                self.SN,
                self.XG,
                self.YG,
                self.tree,
                self.tp,
            )
        elif self.readiness["h"] == "local_cartesian":
            h_rel_tuple = find_rel_h_naive(
                x, y, self.XC, self.YC, self.dX, self.dY, self.CS, self.SN, self.tree
            )
        elif self.readiness["h"] == "rectilinear":
            h_rel_tuple = find_rel_h_rectilinear(x, y, self.lon, self.lat)
        return HRel._make(h_rel_tuple)

    def _find_rel_v(self, z):
        """Find the rel-coord based on vertical center grid using the nearest neighbor scheme."""
        iz, rz, dz, bz = find_rel_nearest(z, self.Z)
        return VRel(iz, rz, dz, bz)

    def _find_rel_v_lin(self, z):
        """Find the rel-coord based on vertical center grid using the 2-point linear scheme."""
        iz, rz, dz, bz = find_rel_z(z, self.Z, self.dZ)
        return VLinRel(iz, rz, dz, bz)

    def _find_rel_vl(self, z):
        """Find the rel-coord based on vertical staggered grid using the nearest neighbor scheme."""
        iz, rz, dz, bz = find_rel_nearest(z, self.Zl)
        return VlRel(iz, rz, dz, bz)

    def _find_rel_vl_lin(self, z):
        """Find the rel-coord based on vertical staggered grid using the 2-point linear scheme."""
        iz, rz, dz, bz = find_rel_z(z, self.Zl, self.dZl, dz_above_z=False)
        return VlLinRel(iz, rz, dz, bz)

    def _find_rel_t(self, t):
        """Find the rel-coord based on the temporal direction using the nearest neighbor scheme."""
        it, rt, dt, bt = find_rel_nearest(t, self.ts)
        return TRel(it, rt, dt, bt)

    def _find_rel_t_lin(self, t):
        """Find the rel-coord based on the temporal direction using the 2-point linear scheme."""
        it, rt, dt, bt = find_rel_time(t, self.ts)
        return TLinRel(it, rt, dt, bt)



import numpy as np
import pytest

import seaduck as sd


@pytest.fixture
def one_p():
    od = sd.OceData(sd.utils.get_dataset("ecco"))
    return sd.Particle(
        x=0.0,
        y=0.0,
        z=-10.0,
        t=sd.utils.convert_time("1992-02-01"),
        data=od,
    )


def not_out_of_bound_in_analytical_step(new_p, tf=1e80, tol=1e-4):
    u_list, du_list, pos_list = new_p._extract_velocity_position()
    tf = np.array([tf])
    ts = sd.lagrangian._time2wall(pos_list, u_list, du_list, tf)
    tend, t_event = sd.lagrangian._which_early(tf, ts)
    new_x, new_u = new_p._move_within_cell(t_event, u_list, du_list, pos_list)
    for rr in new_x:
        try:
            assert not np.logical_or(rr > 0.5 + tol, rr < -0.5 - tol).any()
        except AssertionError:
            where = np.where(np.logical_or(rr > 0.5 + tol, rr < -0.5 - tol))[0][0]
            raise ValueError(
                f"Particle way out of bound."
                # f"tend = {tend[where]},"
                f" t_event = {t_event[where]},"
                f" rx = {new_x[0][where]},ry = {new_x[1][where]},rz = {new_x[2][where]}"
                f"start with u = {new_p.u[where]}, du = {new_p.du[where]}, x={new_p.rx[where]}"
                f"start with v = {new_p.v[where]}, dv = {new_p.dv[where]}, y={new_p.ry[where]}"
                f"start with w = {new_p.w[where]}, dw = {new_p.dv[where]}, z={new_p.rzl_lin[where]}"
            )
    return tend[0]


def random_p(one_p, seed):
    np.random.seed(seed)
    (
        one_p.u,
        one_p.du,
        one_p.rx,
        one_p.v,
        one_p.dv,
        one_p.ry,
        one_p.w,
        one_p.dw,
        one_p.rzl_lin,
    ) = (np.array([i]) for i in np.random.uniform(-0.5, 0.5, 9))

    one_p.rzl_lin += 0.5

    one_p.u /= 1e5
    one_p.v /= 1e5
    one_p.w /= 1e5

    one_p.du /= 1e5
    one_p.dv /= 1e5
    one_p.dw /= 1e5

    return one_p


def u_du_from_uwall(ul, ur, rx):
    u = ul * (0.5 - rx) + ur * (rx + 0.5)
    du = ur - ul
    return u, du


def test_reproduce_issue55(one_p):
    one_p.u[:] = 4.695701621346472e-06
    one_p.du[:] = -5.773571643116384e-05
    one_p.rx[:] = -0.009470161238869457
    one_p.v[:] = -1.4674619307268205e-05
    one_p.dv[:] = -2.8246484204894833e-06
    one_p.ry[:] = 0.08627583529220939
    one_p.w[:] = 0.0
    one_p.dw[:] = -2.8246484204894833e-06
    one_p.rzl_lin[:] = 0.5 + 0.3454546420042159
    tend = not_out_of_bound_in_analytical_step(one_p)
    assert tend != 6


def test_exactly_at_corner(one_p):
    # reproduce_issue55vol2
    u = [0.00023098]
    du = [-1.58748431e-05]
    x = [-0.5]
    v = [-4.54496561e-06]
    dv = [1.31782954e-05]
    y = [-0.5]
    w = [0.0]
    dw = [1.31782954e-05]
    z = [0.34545464]

    one_p.u[:] = np.array(u)
    one_p.du[:] = np.array(du)
    one_p.rx[:] = np.array(x)
    one_p.v[:] = np.array(v)
    one_p.dv[:] = np.array(dv)
    one_p.ry[:] = np.array(y)
    one_p.w[:] = np.array(w)
    one_p.dw[:] = np.array(dw)
    one_p.rzl_lin[:] = 0.5 + np.array(z)
    tend = not_out_of_bound_in_analytical_step(one_p)
    assert tend != 6


@pytest.mark.parametrize("seed", [43, 20, 628])
def test_underflow_du(one_p, seed):
    one_p = random_p(one_p, seed)

    one_p.du /= 1e5
    one_p.dv /= 1e5
    one_p.dw /= 1e5

    tend = not_out_of_bound_in_analytical_step(one_p)
    assert tend != 6


@pytest.mark.parametrize("seed", [432, 320, 60288])
def test_underflow_u(one_p, seed):
    one_p = random_p(one_p, seed)

    one_p.u /= 1e5
    one_p.v /= 1e5
    one_p.w /= 1e5

    tend = not_out_of_bound_in_analytical_step(one_p)
    assert tend != 6


@pytest.mark.parametrize("seed", [0])
def test_u_du_uwall_conversion(one_p, seed):
    one_p = random_p(one_p, seed)
    ul, ur = sd.utils._uleftright_from_udu(one_p.u, one_p.du, one_p.rx)
    u, du = u_du_from_uwall(ul, ur, one_p.rx)
    assert np.allclose(u, one_p.u)
    assert np.allclose(du, one_p.du)


@pytest.mark.parametrize("seed", list(range(1999, 2011)))
def test_impossible_wall(one_p, seed):
    not_wall = np.random.randint(0, 2, 6)
    passages = round(np.sum(not_wall))
    where = list(np.where(not_wall)[0])
    uwall_list = np.zeros(6)
    vels = np.random.uniform(-1e-5, 1e-5, passages)
    vels[-1] = -np.sum(vels[:-1])

    uwall_list[not_wall.astype(bool)] = vels
    uwall_list *= np.array([1, -1, 1, -1, 1, -1])
    uwall_list = [np.array([uu]) for uu in uwall_list]

    one_p = random_p(one_p, seed)
    one_p.u, one_p.du = u_du_from_uwall(uwall_list[0], uwall_list[1], one_p.rx)
    one_p.v, one_p.dv = u_du_from_uwall(uwall_list[2], uwall_list[3], one_p.ry)
    one_p.w, one_p.dw = u_du_from_uwall(
        uwall_list[4], uwall_list[5], one_p.rzl_lin - 0.5
    )

    tend = not_out_of_bound_in_analytical_step(one_p)
    if passages > 1:
        assert tend in where, uwall_list


@pytest.mark.parametrize("seed", list(range(2011, 2023)))
def test_impossible_inflow(one_p, seed):
    not_wall = np.random.randint(0, 2, 6)
    passages = round(np.sum(not_wall))
    where = list(np.where(not_wall)[0])
    uwall_list = np.random.random(6)
    uwall_list[~not_wall.astype(bool)] = 1
    vels = np.random.uniform(-1e-5, 1e-5, passages)
    vels[-1] = -np.sum(vels[:-1])

    uwall_list[not_wall.astype(bool)] = vels
    uwall_list *= np.array([1, -1, 1, -1, 1, -1])
    uwall_list = [np.array([uu]) for uu in uwall_list]

    one_p = random_p(one_p, seed)
    one_p.u, one_p.du = u_du_from_uwall(uwall_list[0], uwall_list[1], one_p.rx)
    one_p.v, one_p.dv = u_du_from_uwall(uwall_list[2], uwall_list[3], one_p.ry)
    one_p.w, one_p.dw = u_du_from_uwall(
        uwall_list[4], uwall_list[5], one_p.rzl_lin - 0.5
    )

    tend = not_out_of_bound_in_analytical_step(one_p)
    if passages > 1:
        assert tend in where, uwall_list


@pytest.mark.parametrize("seed", list(range(2023, 2028)))
@pytest.mark.parametrize("tf", [1e80, -1e80])
def test_cross_cell_wall(one_p, seed, tf):
    one_p = random_p(one_p, seed)
    tend = one_p.analytical_step(tf)
    one_p.cross_cell_wall(tend)
    if (abs(one_p.rx) > 1).any():
        where = np.where(abs(one_p.rx) > 1)[0][0]
        raise ValueError(
            f"lon = {one_p.lon[where]}, lat = {one_p.lat[where]}, "
            f"px = {one_p.px.T[where]}, py = {one_p.py.T[where]}"
        )



import copy

import numpy as np
import xarray as xr

from seaduck.eulerian import Position
from seaduck.runtime_conf import compileable
from seaduck.utils import (
    _time2wall,
    _uleftright_from_udu,
    _which_early,
    parallelpointinpolygon,
)

try:  # pragma: no cover
    import zarr
except ImportError:  # pragma: no cover
    pass

UV_DIC = {"U": 0, "V": 1}
MOVE_DIC = {
    0: np.array([0, 0]),  # delta_y,delta_x
    1: np.array([0, 1]),
    2: np.array([0, 0]),
    3: np.array([1, 0]),
    4: np.array([0, 0]),
    5: np.array([0, 0]),
}


def read_from_ds(particle_ds, oce):
    temp = Position.__new__(Position)
    temp.ocedata = oce
    temp.tp = temp.ocedata.tp

    # it = np.array(particle_ds.it)
    if oce.tp.typ in ["LLC"]:
        temp.face = np.array(particle_ds.fc).astype(int)
    else:
        temp.face = None
    izl = np.array(particle_ds.iz)
    iy = np.array(particle_ds.iy)
    ix = np.array(particle_ds.ix)
    rzl = np.array(particle_ds.rz)
    ry = np.array(particle_ds.ry)
    rx = np.array(particle_ds.rx)

    # temp.it  = it .astype(int)
    temp.izl_lin = izl.astype(int)
    temp.iz = (izl - 1).astype(int)
    temp.iy = iy.astype(int)
    temp.ix = ix.astype(int)
    temp.rzl_lin = rzl
    temp.ry = ry
    temp.rx = rx

    temp.N = len(temp.ix)

    uu = np.array(particle_ds.uu)
    vv = np.array(particle_ds.vv)
    ww = np.array(particle_ds.ww)
    du = np.array(particle_ds.du)
    dv = np.array(particle_ds.dv)
    dw = np.array(particle_ds.dw)

    temp.u = uu
    temp.v = vv
    temp.w = ww
    temp.du = du
    temp.dv = dv
    temp.dw = dw

    temp.lon = np.array(particle_ds.xx)
    temp.lat = np.array(particle_ds.yy)
    temp.dep = np.array(particle_ds.zz)
    temp.t = np.array(particle_ds.tt)

    temp.vs = np.array(particle_ds.vs)

    temp.shapes = list(particle_ds.shapes)

    return temp


def which_wall(pt):
    distl = np.abs(0.5 + pt.rx)
    distr = np.abs(0.5 - pt.rx)
    distd = np.abs(0.5 + pt.ry)
    distt = np.abs(0.5 - pt.ry)
    distdeep = np.abs(pt.rzl_lin)
    distshal = np.abs(1 - pt.rzl_lin)

    distance = np.vstack([distl, distr, distd, distt, distdeep, distshal])
    return np.argmin(distance, axis=0)


def pseudo_motion(pt):
    us = [pt.u, pt.v, pt.w]
    dus = [pt.du, pt.dv, pt.dw]
    xs = [pt.rx, pt.ry, pt.rzl_lin - 1 / 2]
    ts_forward = _time2wall(xs, us, dus, 1e80 * np.ones(pt.N))
    ts_backward = _time2wall(xs, us, dus, -1e80 * np.ones(pt.N))
    tendf, tf = _which_early(1e80, ts_forward)
    tendb, tb = _which_early(-1e80, ts_backward)

    return tendf, tf, tendb, tb


@compileable
def fast_cumsum(shapes):
    return np.cumsum(shapes)


def first_last_neither(shapes, return_neither=True):
    acc = fast_cumsum(shapes)
    last = acc - 1
    first = np.roll(acc, 1)
    first[0] = 0
    if not return_neither:
        return first, last
    else:
        neither = np.array(
            [
                first[i] + j
                for i, length in enumerate(shapes)
                for j in range(1, length - 1)
            ]
        )
        return first, last, neither


def pt_ulist(pt):
    ul, ur = _uleftright_from_udu(pt.u, pt.du, pt.rx)
    vl, vr = _uleftright_from_udu(pt.v, pt.dv, pt.ry)
    wl, wr = _uleftright_from_udu(pt.w, pt.dw, pt.rzl_lin - 0.5)
    return np.array([ul, ur, vl, vr, wl, wr])


def residence_time(pt):
    u_list = pt_ulist(pt)
    return 2 / np.sum(np.abs(u_list), axis=0)


def tres_update(tres0, temp, first, last, fraction_first, fraction_last):
    fracs = np.ones(temp.N)
    # fracs_a = np.ones(temp.N)
    fracs[first + 1] = fraction_first
    # fracs_a[first+1] = fraction_first
    fracs[last] -= fraction_last
    tres = tres0 * fracs

    tres[temp.vs > 6] = 0.0

    return tres


def tres_fraction(temp, first, last, fraction_first, fraction_last):
    tres0 = residence_time(temp)
    tres = tres_update(tres0, temp, first, last, fraction_first, fraction_last)
    return tres


def ind_tend_uv(ind, tp):
    """Return the index of the velocity node.

    The node is not of the same index as the center point.
    iw determines where to return iw
    """
    iw, face, iy, ix = ind
    if iw == 1:
        new_wall, new_ind = tp._ind_tend_V((face, iy, ix), 0)
    elif iw == 0:
        new_wall, new_ind = tp._ind_tend_U((face, iy, ix), 3)
    else:
        raise ValueError("illegal iw")
    niw = UV_DIC[new_wall]
    nfc, niy, nix = new_ind
    return niw, nfc, niy, nix


def deepcopy_inds(temp):
    iz = copy.deepcopy(temp.izl_lin)
    iy = copy.deepcopy(temp.iy)
    ix = copy.deepcopy(temp.ix)
    if temp.face is not None:
        face = copy.deepcopy(temp.face)
        return iz, face, iy, ix
    else:
        return iz, iy, ix


def wall_index(inds, iwall, tp):
    iw = iwall // 2
    iz = copy.deepcopy(inds[0])
    iy = copy.deepcopy(inds[-2])
    ix = copy.deepcopy(inds[-1])

    ind = np.array(inds[1:])
    old_ind = copy.deepcopy(ind)
    naive_move = np.array([MOVE_DIC[i] for i in iwall], dtype=int).T
    ind[-2] += naive_move[0]  # iy
    iy += naive_move[0]
    ind[-1] += naive_move[1]  # ix
    ix += naive_move[1]

    iz[iwall == 4] += 1
    iz -= 1
    if tp.typ in ["LLC"]:
        face = copy.deepcopy(inds[-3])
        illegal = tp.check_illegal(ind, cuvwg="G")
        redo = np.array(np.where(illegal)).T
        for num, loc in enumerate(redo):
            j = loc[0]
            ind = (iw[j],) + tuple(old_ind[:, j])
            new_ind = ind_tend_uv(ind, tp)
            iw[j], face[j], iy[j], ix[j] = new_ind

        return np.array([iw, iz, face, iy, ix]).astype(int)
    else:
        return np.array([iw, iz, ind[-2], ind[-1]]).astype(int)


def redo_index(pt):
    inds = deepcopy_inds(pt)
    tendf, tf, tendb, tb = pseudo_motion(pt)

    funderflow = np.where(tendf == 6)
    bunderflow = np.where(tendb == 6)
    tendf[funderflow] = 0
    tendb[bunderflow] = 0
    vf = wall_index(inds, tendf, pt.ocedata.tp)
    vb = wall_index(inds, tendb, pt.ocedata.tp)
    tim = tf - tb
    frac = -tb / tim
    assert (~np.isnan(tim)).any(), [
        i[np.isnan(tim)] for i in [pt.rx, pt.ry, pt.rzl_lin - 1 / 2]
    ]
    # assert (tim != 0).all(), [i[tim == 0] for i in [pt.rx, pt.ry, pt.rzl_lin - 1 / 2]]
    # at_corner = np.where(tim == 0)
    # frac[at_corner] = 1
    frac = np.nan_to_num(frac, nan=1)
    return vf, vb, frac


def find_ind_frac_tres(neo, oce, region_names=False, region_polys=None, by_type=True):
    temp = read_from_ds(neo, oce)
    temp.shapes = list(temp.shapes)
    if region_names:
        masks = []
        for reg in region_polys:
            mask = parallelpointinpolygon(temp.lon, temp.lat, reg)
            # mask = np.where(mask)[0]
            masks.append(mask)

    first, last, neither = first_last_neither(np.array(temp.shapes))
    if temp.face is not None:
        num_ind = 5
    else:
        num_ind = 4

    if by_type:
        ind1 = np.zeros((num_ind, temp.N), "int16")
        ind2 = np.ones((num_ind, temp.N), "int16")
        frac = np.ones(temp.N)

        # ind1[:, wrong_ind] = lookup[:, lookup_ind]

        if len(neither > 0):
            neithers = temp.subset(neither)
            neither_inds = deepcopy_inds(neithers)
            iwalls = which_wall(neithers)
            ind1[:, neither] = wall_index(neither_inds, iwalls, temp.ocedata.tp)

        firsts = temp.subset(first)
        lasts = temp.subset(last)
        ind1[:, first], ind2[:, first], frac[first] = redo_index(firsts)
        ind1[:, last], ind2[:, last], frac[last] = redo_index(lasts)
    else:
        ind1, ind2, frac = redo_index(temp)

    tres = tres_fraction(temp, first, last, frac[first], frac[last])
    if region_names:
        return ind1, ind2, frac, masks, tres, last, first
    else:
        return ind1, ind2, frac, tres, last, first


def flatten(lstoflst, shapes=None):
    if shapes is None:
        shapes = [len(i) for i in lstoflst]
    suffix = np.cumsum(shapes)
    thething = np.zeros(suffix[-1])
    thething[: suffix[0]] = lstoflst[0]
    for i in range(1, len(lstoflst)):
        thething[suffix[i - 1] : suffix[i]] = lstoflst[i]
    return thething


def particle2xarray(p):
    shapes = [len(i) for i in p.xxlist]
    # it = flatten(p.itlist,shapes = shapes)
    if p.face is not None:
        fc = flatten(p.fclist, shapes=shapes)
    iy = flatten(p.iylist, shapes=shapes)
    iz = flatten(p.izlist, shapes=shapes)
    ix = flatten(p.ixlist, shapes=shapes)
    rx = flatten(p.rxlist, shapes=shapes)
    ry = flatten(p.rylist, shapes=shapes)
    rz = flatten(p.rzlist, shapes=shapes)
    tt = flatten(p.ttlist, shapes=shapes)
    uu = flatten(p.uulist, shapes=shapes)
    vv = flatten(p.vvlist, shapes=shapes)
    ww = flatten(p.wwlist, shapes=shapes)
    du = flatten(p.dulist, shapes=shapes)
    dv = flatten(p.dvlist, shapes=shapes)
    dw = flatten(p.dwlist, shapes=shapes)
    xx = flatten(p.xxlist, shapes=shapes)
    yy = flatten(p.yylist, shapes=shapes)
    zz = flatten(p.zzlist, shapes=shapes)
    vs = flatten(p.vslist, shapes=shapes)

    ds = xr.Dataset(
        coords=dict(shapes=(["shapes"], shapes), nprof=(["nprof"], np.arange(len(xx)))),
        data_vars=dict(
            # it = (['nprof'],it),
            iy=(["nprof"], iy),
            iz=(["nprof"], iz),
            ix=(["nprof"], ix),
            rx=(["nprof"], rx),
            ry=(["nprof"], ry),
            rz=(["nprof"], rz),
            tt=(["nprof"], tt),
            uu=(["nprof"], uu),
            vv=(["nprof"], vv),
            ww=(["nprof"], ww),
            du=(["nprof"], du),
            dv=(["nprof"], dv),
            dw=(["nprof"], dw),
            xx=(["nprof"], xx),
            yy=(["nprof"], yy),
            zz=(["nprof"], zz),
            vs=(["nprof"], vs),
        ),
    )
    if p.face is not None:
        ds["fc"] = xr.DataArray(fc, dims="nprof")
    return ds


def dump_to_zarr(
    neo, oce, filename, region_names=False, region_polys=None, preserve_checks=False
):
    if region_names:
        (ind1, ind2, frac, masks, tres, last, first) = find_ind_frac_tres(
            neo, oce, region_names=region_names, region_polys=region_polys
        )
    else:
        ind1, ind2, frac, tres, last, first = find_ind_frac_tres(neo, oce)

    if oce.tp.typ in ["LLC"]:
        neo["face"] = neo["fc"].astype("int16")
        neo["five"] = xr.DataArray(["iw", "iz", "face", "iy", "ix"], dims="five")
    else:
        neo["five"] = xr.DataArray(["iw", "iz", "iy", "ix"], dims="five")
    if region_names:
        for ir, reg in enumerate(region_names):
            neo[reg] = xr.DataArray(masks[ir].astype(bool), dims="nprof")
        # neo['gulf'] = xr.DataArray(gulf_ind.astype(bool),dims = 'nprof')
        # neo['labr'] = xr.DataArray(labr_ind.astype(bool),dims = 'nprof')
        # neo['gdbk'] = xr.DataArray(gdbk_ind.astype(bool),dims = 'nprof')
        # neo['nace'] = xr.DataArray(nace_ind.astype(bool),dims = 'nprof')
        # neo['egrl'] = xr.DataArray(egrl_ind.astype(bool),dims = 'nprof')
        # region_ind = np.concatenate([gulf_ind,labr_ind,gdbk_ind,nace_ind,egrl_ind]).astype('int32')
        # neo.attrs['region_shape'] = [len(i) for i in [gulf_ind,labr_ind,gdbk_ind,nace_ind,egrl_ind]]
        # if len(region_ind)>0:
        #     neo = neo.assign_coords(region_ind = xr.DataArray(region_ind,dims = 'region_ind'))

    neo["ind1"] = xr.DataArray(ind1.astype("int16"), dims=["five", "nprof"])
    neo["ind2"] = xr.DataArray(ind2.astype("int16"), dims=["five", "nprof"])
    neo["frac"] = xr.DataArray(frac, dims="nprof")
    neo["tres"] = xr.DataArray(tres, dims="nprof")
    # neo['last'] = xr.DataArray(last.astype('int64'), dims = 'shapes')
    # neo['first'] = xr.DataArray(first.astype('int64'), dims = 'shapes')

    neo["ix"] = neo["ix"].astype("int16")
    neo["iy"] = neo["iy"].astype("int16")
    neo["iz"] = neo["iz"].astype("int16")
    neo["vs"] = neo["vs"].astype("int16")

    if not preserve_checks:
        neo = neo.drop_vars(["rx", "ry", "rz", "uu", "vv", "ww", "du", "dv", "dw"])
    if "fc" in neo.data_vars:
        neo = neo.drop_vars(["fc"])

    neo.to_zarr(filename, mode="w")
    zarr.consolidate_metadata(filename)


def store_lists(pt, name, region_names=False, region_polys=None, **kwarg):
    neo = particle2xarray(pt)
    dump_to_zarr(
        neo,
        pt.ocedata,
        name,
        region_names=region_names,
        region_polys=region_polys,
        **kwarg
    )


def prefetch_scalar(ds_slc, scalar_names):
    prefetch = {}
    for var in scalar_names:
        # print(var, end = ' ')
        prefetch[var] = np.array(ds_slc[var])
    return prefetch


def read_wall_list(neo, tp, prefetch=None, scalar=True):
    if "face" not in neo.data_vars:
        ind = (neo.iz - 1, neo.iy, neo.ix)
        deep_ind = (neo.iz, neo.iy, neo.ix)
        right_ind = tuple(
            [neo.iz - 1]
            + list(
                tp.ind_tend_vec(
                    (neo.iy, neo.ix), np.ones(len(neo.nprof)) * 3, cuvwg="G"
                )
            )
        )
        up_ind = tuple(
            [neo.iz - 1]
            + list(tp.ind_tend_vec((neo.iy, neo.ix), np.ones(len(neo.nprof)) * 0))
        )
        uarray, varray, warray = prefetch
        ur = uarray[right_ind]
        vr = varray[up_ind]
    else:
        ind = (neo.iz - 1, neo.face, neo.iy, neo.ix)
        deep_ind = (neo.iz, neo.face, neo.iy, neo.ix)
        right_ind = tuple(
            [neo.iz - 1]
            + list(
                tp.ind_tend_vec((neo.face, neo.iy, neo.ix), np.ones(len(neo.nprof)) * 3)
            )
        )
        up_ind = tuple(
            [neo.iz - 1]
            + list(
                tp.ind_tend_vec((neo.face, neo.iy, neo.ix), np.ones(len(neo.nprof)) * 0)
            )
        )
        uarray, varray, warray = prefetch

        ur_temp = np.nan_to_num(uarray[right_ind])
        vr_temp = np.nan_to_num(varray[right_ind])
        uu_temp = np.nan_to_num(uarray[up_ind])
        vu_temp = np.nan_to_num(varray[up_ind])
        right_faces = np.vstack([ind[1], right_ind[1]]).T
        up_faces = np.vstack([ind[1], up_ind[1]]).T
        ufromu, ufromv, _, _ = tp.four_matrix_for_uv(right_faces)
        _, _, vfromu, vfromv = tp.four_matrix_for_uv(up_faces)
        if scalar:
            ufromu, ufromv, vfromu, vfromv = (
                np.abs(i) for i in [ufromu, ufromv, vfromu, vfromv]
            )

        ur = ur_temp * ufromu[:, 1] + vr_temp * ufromv[:, 1]
        vr = uu_temp * vfromu[:, 1] + vu_temp * vfromv[:, 1]
    ul = uarray[ind]
    vl = varray[ind]
    wr = warray[ind]
    wl = warray[deep_ind]
    return np.array([np.nan_to_num(i) for i in (ul, ur, vl, vr, wl, wr)])


def crude_convergence(u_list):
    conv = np.array(u_list).T * np.array([1, -1, 1, -1, 1, -1])
    conv = np.sum(conv, axis=-1)
    return conv


def check_particle_data_compat(
    xrpt,
    xrslc,
    tp,
    use_tracer_name=None,
    wall_names=("sx", "sy", "sz"),
    conv_name="divus",
    debug=False,
):
    if "iz" not in xrpt.data_vars:
        raise NotImplementedError(
            "This functionality only support 3D simulation at the moment."
        )
    if isinstance(use_tracer_name, str):
        wall_names = tuple(use_tracer_name + i for i in ["x", "y", "z"])
        conv_name = "divu" + use_tracer_name
    elif use_tracer_name is not None:
        raise ValueError("use_tracer_name has to be a string.")
    prefetch = []
    for var in wall_names:
        prefetch.append(np.array(xrslc[var]))
    c_list = read_wall_list(xrpt, tp, prefetch)

    ul, ur = _uleftright_from_udu(xrpt.uu, xrpt.du, xrpt.rx)
    vl, vr = _uleftright_from_udu(xrpt.vv, xrpt.dv, xrpt.ry)
    wl, wr = _uleftright_from_udu(xrpt.ww, xrpt.dw, xrpt.rz - 0.5)
    u_list = np.array([np.array(i) for i in [ul, ur, vl, vr, wl, wr]])

    flux_list = c_list * u_list
    lagrangian_conv = crude_convergence(flux_list)

    if "face" in xrpt.data_vars:
        ind = (xrpt.iz - 1, xrpt.face, xrpt.iy, xrpt.ix)
    else:
        ind = (xrpt.iz - 1, xrpt.iy, xrpt.ix)
    eulerian_conv = np.array(xrslc[conv_name])[ind]
    if debug:
        extra = (u_list, c_list, lagrangian_conv, eulerian_conv)
    else:
        extra = None
    return np.allclose(lagrangian_conv, eulerian_conv), extra


def prefetch_vector(
    ds_slc, xname="sxprime", yname="syprime", zname="szprime", same_size=True
):
    if same_size:
        return np.array(tuple(np.array(ds_slc[i]) for i in [xname, yname, zname]))
    else:
        xx = np.array(ds_slc[xname])
        yy = np.array(ds_slc[yname])
        zz = np.array(ds_slc[zname])
        shape = (3,) + tuple(
            int(np.max([ar.shape[j] for ar in [xx, yy, zz]]))
            for j in range(len(xx.shape))
        )
        larger = np.empty(shape)
        larger[(0,) + tuple(slice(i) for i in xx.shape)] = xx
        larger[(1,) + tuple(slice(i) for i in yy.shape)] = yy
        larger[(2,) + tuple(slice(i) for i in zz.shape)] = zz
        return larger


def read_prefetched_scalar(ind, scalar_names, prefetch):
    res = {}
    for var in scalar_names:
        res[var] = prefetch[var][ind]
    return res


def lhs_contribution(t, scalar_dic, last, lhs_name="lhs"):
    deltat = np.nan_to_num(np.diff(t))
    deltat[last[:-1]] = 0
    lhs_scalar = scalar_dic[lhs_name][:-1]
    correction = deltat * lhs_scalar
    return correction


def contr_p_relaxed(deltas, tres, step_dic, termlist, p=1, error_prefix=""):
    nds = len(deltas)
    # if len(wrong_ind)>0:
    #     if wrong_ind[-1] == len(deltas):
    #         wrong_ind = wrong_ind[:-1]
    dic = {error_prefix + "error": np.zeros(nds)}
    # dic['error'][wrong_ind] = deltas[wrong_ind]
    # deltas[wrong_ind] = 0
    # tres[wrong_ind] = 0

    deno = np.zeros(nds)
    sums = np.zeros(nds)
    for var in termlist:
        deno += step_dic[var][:-1] ** (p + 1)
        sums += step_dic[var][:-1]
    disparity = deltas - sums * tres
    total = np.zeros(nds)
    # dic['quality'] = np.nan_to_num(np.abs((disparity/tres)**(p+1)/deno))
    # mask = (dic['quality']<=1).astype(int)
    for var in termlist:
        ratio = step_dic[var][:-1] ** (p + 1) / deno
        dic[var] = step_dic[var][:-1] * tres + ratio * disparity
        total += dic[var]
    final_correction = deltas - total
    # assert np.allclose(final_correction, 0)
    dic[error_prefix + "error"] += final_correction
    return dic



import numpy as np
import pytest
import xarray as xr

from seaduck.smart_read import smart_read


def old_smart_read(da, ind, memory_chunk=3, dask_more_efficient=100):
    """Read from a xarray.DataArray given tuple indexes.

    Try to do it fast and smartly.

    Parameters
    ----------
    da: xarray.DataArray
        DataArray to read from
    ind: tuple of numpy.ndarray
        The indexes of points of interest, each element does not need to be 1D
    memory_chunk: int, default 3
        If the number of chunks needed is smaller than this, read all of them at once.
    xarray_more_efficient: int, default 100
        When the number of chunks is larger than this, and the data points are few,
        it may make sense to directly use xarray's vectorized read.

    Returns
    -------
    + values: numpy.ndarray
        The values of the points of interest. Has the same shape as the elements in ind.
    """
    the_shape = ind[0].shape
    ind = tuple(i.ravel() for i in ind)
    if len(da.dims) != len(ind):
        raise ValueError("index does not match the number of dimensions")
    if da.chunks is None or da.chunks == {}:
        npck = np.array(da)
        return npck[ind].reshape(the_shape)
    if (
        np.prod([len(i) for i in da.chunks]) <= memory_chunk
    ):  # if the number of chunks is small don't bother
        npck = np.array(da)
        return npck[ind].reshape(the_shape)
    cksz = dict(zip(da.dims, da.chunks))
    keys = list(cksz.keys())
    n = len(ind[0])
    result = np.zeros(n)

    new_dic = {}
    # typically what happens is that the first a few indexes are chunked
    # here we figure out what is the last dimension chunked.
    for i in range(len(cksz) - 1, -1, -1):
        if len(cksz[keys[i]]) > 1:
            last = i
            break

    ckbl = np.zeros((n, i + 1)).astype(int)
    # register each each dimension and the chunk they are in
    for i, k in enumerate(keys[: i + 1]):
        ix = ind[i]
        suffix = np.cumsum(cksz[k])
        new_dic[i] = suffix
        ckbl[:, i] = np.searchsorted(suffix, ix, side="right")
    # this is the time limiting step for localized long query.
    ckus, inverse = np.unique(ckbl, axis=0, return_inverse=True)
    # ckus is the individual chunks used
    if len(ckus) <= dask_more_efficient:
        # logging.debug('use smart')
        for i, k in enumerate(ckus):
            ind_str = []
            pre = []
            which = (inverse == i).ravel()
            for j, p in enumerate(k):
                sf = new_dic[j][p]  # the upperbound of index
                pr = sf - cksz[keys[j]][p]  # the lower bound of index
                ind_str.append(slice(pr, sf))
                pre.append(pr)
            prs = np.zeros(len(keys)).astype(int)
            prs[: last + 1] = pre
            npck = np.array(da[tuple(ind_str)])
            subind = tuple(ind[dim][which] - prs[dim] for dim in range(len(ind)))
            result[which] = npck[subind]
        return result.reshape(the_shape)
    else:
        # logging.debug('use xarray')
        xrind = tuple(xr.DataArray(dim, dims=["x"]) for dim in ind)
        return np.array(da[xrind]).reshape(the_shape)


@pytest.fixture
def ind():
    N = 3
    it = np.arange(N).astype(int)
    iz = np.arange(N).astype(int)
    fc = np.arange(N).astype(int)
    ix = np.arange(N).astype(int)
    iy = np.arange(N, 2 * N).astype(int)
    return (it, iz, fc, iy, ix)


@pytest.mark.parametrize("chunk", [{"time": 1}, {"time": 1, "Z": 1}, {}])
@pytest.mark.parametrize("ds", ["ecco"], indirect=True)
def test_just_read(ind, ds, chunk):
    ds["SALT"] = ds["SALT"].chunk(chunk)
    assert np.allclose(smart_read(ds["SALT"], ind), old_smart_read(ds["SALT"], ind))


@pytest.mark.parametrize("ds", ["ecco"], indirect=True)
def test_read_xarray(ind, ds):
    ds["SALT"] = ds["SALT"].chunk({"time": 1})
    assert np.allclose(
        smart_read(ds["SALT"], ind, dask_more_efficient=1),
        old_smart_read(ds["SALT"], ind, dask_more_efficient=1),
    )


@pytest.mark.parametrize("ds", ["ecco"], indirect=True)
def test_mismatch_read(ind, ds):
    with pytest.raises(ValueError):
        smart_read(ds["XC"], ind)



import numpy as np
import pytest
import xarray as xr
from scipy import spatial

import seaduck as sd


def test_chg_ref_lon():
    assert sd.utils.chg_ref_lon(-5, 125) == 230


@pytest.mark.parametrize("item,length", [([1, 2, 3], 3), (0, 1)])
def test_general_len(item, length):
    assert sd.utils._general_len(item) == length


def test_to_180():
    assert sd.utils.to_180(365) == 5


def test_to_180_underflow():
    assert abs(sd.utils.to_180(-1e-20)) < 1e-10


def test_key_by_values():
    dic = {1: "a", "hello": "b", "alpha": "a"}
    assert sd.utils.get_key_by_value(dic, "b") == "hello"
    assert sd.utils.get_key_by_value(dic, "a") in [1, "alpha"]
    assert sd.utils.get_key_by_value(dic, "c") is None


@pytest.mark.parametrize(
    "lat,lon",
    [
        (0, 0),
        (45, 45),
        (50, 180),
    ],
)
def test_spherical2cartesian(lat, lon):
    X, Y, Z = sd.utils.spherical2cartesian(lon, lat)
    assert np.isclose(X**2 + Y**2 + Z**2, 6371.0**2)


def test_local_to_latlon():
    u = 1
    v = 0.618
    cs = np.random.random()
    sn = np.sqrt(1 - cs**2)
    uu, vv = sd.utils.local_to_latlon(u, v, cs, sn)
    assert np.isclose(np.hypot(uu, vv), np.hypot(u, v))


def test_none_in():
    assert sd.utils.NoneIn([1, 2, None])
    assert not (sd.utils.NoneIn([1, 2, 3]))


@pytest.mark.parametrize("ds", ["curv"], indirect=True)
def test_cs_sn(ds):
    cs, sn = sd.utils.missing_cs_sn(ds, return_xr=True)
    assert isinstance(cs, xr.DataArray)


def test_covert_time():
    rand = round(np.random.random() * 1e7)
    time = np.datetime64("1970-01-01") + np.timedelta64(rand, "s")
    ress = sd.utils.convert_time(time)
    assert np.isclose(rand, ress, atol=1)


def test_easy_cube():
    east = -80.0
    west = 0.0
    south = 40.0
    north = 75.0
    shallow = -10.0
    deep = -10.0
    time_string = "1977-01-01"

    Nlon = 10
    Nlat = 10
    Ndep = 1

    x, y, z, t = sd.utils.easy_3d_cube(
        (east, west, Nlon),
        (south, north, Nlat),
        (shallow, deep, Ndep),
        time_string,
        print_total_number=True,
    )
    assert len(x) == Nlon * Nlat * Ndep


ascend = np.array([1, 11, 111, 1111])
dscend = -ascend
darray = np.array([1, 10, 100, 1000])


@pytest.mark.parametrize(
    "value,array,ascending,above,peri,ans",
    [
        (10, ascend, 1, True, None, 0),
        (10, ascend, 1, False, None, 1),
        (0, ascend, 1, False, None, 0),
        (-2, dscend, -1, True, None, 1),
        (-2, dscend, -1, False, None, 0),
        (0, ascend, 1, False, 110.5, 2),
        (0, ascend, 1, False, 1111.5, 3),
    ],
)
def test_find_ind(value, array, ascending, above, peri, ans):
    ix, bx = sd.utils.find_ind(
        array, value, peri=peri, ascending=ascending, above=above
    )
    assert ix == ans
    assert isinstance(ix, int)


@pytest.mark.parametrize(
    "value, array, ascending",
    [(0.5, ascend, 1), (-2000, dscend, -1)],
)
def test_find_ind_out_of_bound(value, array, ascending):
    with pytest.raises(ValueError):
        ix, bx = sd.utils.find_ind(array, value, ascending=ascending, above=True)


@pytest.mark.parametrize(
    "value,array,darray,ascending,above,peri, dx_right,ans",
    [
        (10, ascend, darray, 1, True, None, False, 0.9),
        (10, ascend, 10 * darray, 1, True, None, True, 0.9),
        (10, ascend, None, 1, True, None, True, 0.9),
        (10, ascend, darray, 1, False, None, False, -0.1),
        (0, ascend, darray, 1, False, None, False, -1),
        (-2, dscend, darray, -1, True, None, False, 0.9),
        (-2, dscend, darray, -1, False, None, False, -0.1),
        (-2, dscend, 10 * darray, -1, False, None, True, -0.1),
        (0, dscend, darray, -1, True, None, False, 1),
        (0, ascend, darray, 1, False, 110.5, False, -0.5 / 100),
        (0, ascend, 10 * darray, 1, True, 1111.5, True, 0.5 / 10000),
    ],
)
def test_find_rel(value, array, darray, ascending, above, peri, dx_right, ans):
    value = np.array([value])
    ixs, rxs, dxs, bxs = sd.utils.find_rel(
        value, array, darray, ascending, above, peri, dx_right
    )
    assert rxs[0] == ans
    assert np.issubdtype(ixs.dtype, int)


@pytest.mark.parametrize("ds", ["curv", "ecco"], indirect=True)
def test_create_tree_cartesian(ds):
    tree = sd.utils.create_tree(ds["XC"], ds["YC"], R=None)
    assert isinstance(tree, spatial.cKDTree)



import functools
import os

import numpy as np

# Required dependencies (private)
import xarray as xr
from scipy import spatial

from seaduck.runtime_conf import compileable, compileable_parallel

try:
    import numba
except ImportError:
    pass

try:
    import pooch
except ImportError:
    pass


@functools.cache
def pooch_prepare():
    """Prepare for loading datasets using pooch."""
    pooch_testdata = pooch.create(
        path=pooch.os_cache("seaduck"),
        base_url="doi:10.5281/zenodo.7949168",
        registry=None,
    )
    pooch_testdata.load_registry_from_doi()  # Automatically populate the registry
    pooch_fetch_kwargs = {"progressbar": True}
    return pooch_testdata, pooch_fetch_kwargs


def _flower_func(x, y, pedals=8, sin=True, vmin=0, vmax=1):
    theta = np.arctan2(x, y)
    r = np.hypot(x, y)
    center = np.exp(-(r**2) * 8)
    if sin:
        pedal = np.abs(np.sin(pedals / 2 * theta)) / (1 + r)
    else:
        pedal = np.abs(np.cos(pedals / 2 * theta)) / (1 + r)
    return (vmax - vmin) * np.maximum(pedal, center) + vmin


def process_ecco(ds):
    """Add more meat to ECCO dataset after the skeleton is downloaded."""
    np.random.seed(90)
    rand1 = np.random.random((50, 13, 90, 90))
    rand2 = np.random.random((50, 13, 90, 90))
    flower_x, flower_y = tuple(np.linspace(-1, 1, 90) for i in range(2))
    flower_x, flower_y = np.meshgrid(flower_x, flower_y)
    salt_flower = _flower_func(flower_x, flower_y, vmin=33, vmax=36)
    etan_flower = _flower_func(flower_x, flower_y, vmin=-0.1, vmax=0.1, sin=False)
    salt_4d = np.zeros((50, 13, 90, 90))
    salt_4d[:] = salt_flower
    etan_3d = np.zeros((13, 90, 90))
    etan_3d[:] = etan_flower

    ds["UVELMASS"] = xr.DataArray(
        np.stack([rand1, rand2, rand2], axis=0), dims=("time", "Z", "face", "Y", "Xp1")
    )
    ds["UVELMASS"][0] = ds.UVELMASS1
    ds["UVELMASS"][1] = ds.UVELMASS1 * rand1
    ds["UVELMASS"][2] = ds.UVELMASS1 * rand2

    ds["WVELMASS"] = xr.DataArray(
        np.stack([rand1, rand2, rand2], axis=0), dims=("time", "Zl", "face", "Y", "X")
    )
    ds["WVELMASS"][0] = ds.WVELMASS1
    ds["WVELMASS"][1] = ds.WVELMASS1 * rand1
    ds["WVELMASS"][2] = ds.WVELMASS1 * rand2

    ds["VVELMASS"] = xr.DataArray(
        np.stack([rand1, rand2, rand2], axis=0), dims=("time", "Z", "face", "Yp1", "X")
    )
    ds["VVELMASS"][0] = ds.VVELMASS1
    ds["VVELMASS"][1] = ds.VVELMASS1 * rand1
    ds["VVELMASS"][2] = ds.VVELMASS1 * rand2

    ds["SALT"] = xr.DataArray(
        np.stack([salt_4d, salt_4d * 2 - 34.5, salt_4d], axis=0),
        dims=("time", "Z", "face", "Y", "X"),
    )
    ds["SALT_snap"] = xr.DataArray(
        np.stack([salt_4d * 2 - 34.5, salt_4d * 2.5], axis=0),
        dims=("time_midp", "Z", "face", "Y", "X"),
    )
    ds["ETAN"] = xr.DataArray(
        np.stack([etan_3d, etan_3d * 2, etan_3d * 3], axis=0),
        dims=("time", "face", "Y", "X"),
    )
    ds["ETAN_snap"] = xr.DataArray(
        np.stack([etan_3d * 1.5, etan_3d * 2.5], axis=0),
        dims=("time_midp", "face", "Y", "X"),
    )
    ds["utrans"] = ds["UVELMASS"] * ds["drF"] * ds["dyG"]
    ds["vtrans"] = ds["VVELMASS"] * ds["drF"] * ds["dxG"]
    ds["wtrans"] = ds["WVELMASS"] * ds["rA"]
    return ds


@functools.cache
def get_dataset(name):
    """Use pooch to download datasets from cloud.

        This is just for testing purposes.

    Parameters
    ----------
    name: string
        The name of dataset, now support "ecco", "aviso", "curv", "rect"
    """
    pooch_testdata, pooch_fetch_kwargs = pooch_prepare()
    fnames = pooch_testdata.fetch(f"{name}.tar.gz", pooch.Untar(), **pooch_fetch_kwargs)
    ds = xr.open_zarr(os.path.commonpath(fnames))
    if name == "ecco":
        return process_ecco(ds)
    return ds


def chg_ref_lon(x, ref_lon):
    """Change the definition of 0 longitude.

    Return how much east one need to go from ref_lon to x
    This function aims to address
    the confusion caused by the discontinuity in longitude.
    """
    return (x - ref_lon) % 360


def _general_len(thing):
    """Change the definition of len such that the length of a float is 1."""
    try:
        return len(thing)
    except TypeError:
        return 1


def get_key_by_value(d, value):
    """Find one of the keys in a dictionary.

    the key that correspond to the given value.

    Parameters
    ----------
    d: dictionaty
        dictionary to lookup key from
    value: object
        A object that has __eq__ method.
    """
    for k, v in d.items():
        if v == value:
            return k
    return None


@compileable
def spherical2cartesian(lat, lon, R=6371.0):
    """Convert spherical coordinates to cartesian.

    Parameters
    ----------
    lat: np.array
        Spherical Y coordinate (latitude)
    lon: np.array
        Spherical X coordinate (longitude)
    R: scalar
        Earth radius in km
        If None, use geopy default

    Returns
    -------
    x: np.array
        Cartesian x coordinate
    y: np.array
        Cartesian y coordinate
    z: np.array
        Cartesian z coordinate
    """
    # Convert
    y_rad = np.deg2rad(lat)
    x_rad = np.deg2rad(lon)
    x = R * np.cos(y_rad) * np.cos(x_rad)
    y = R * np.cos(y_rad) * np.sin(x_rad)
    z = R * np.sin(y_rad)

    return x, y, z


@compileable
def to_180(x, peri=360):
    """Convert any longitude scale to [-180,180)."""
    x = x % peri % peri
    # underflow resistant
    return x + (-1) * (x // (peri / 2)) * peri


def local_to_latlon(u, v, cs, sn):
    """Convert local vector to north-east."""
    uu = u * cs - v * sn
    vv = u * sn + v * cs
    return uu, vv


@compileable
def rel2latlon(rx, ry, cs, sn, dx, dy, bx, by):
    """Translate the spatial rel-coords into lat-lon-dep coords."""
    temp_x = rx * dx / deg2m
    temp_y = ry * dy / deg2m
    dlon = (temp_x * cs - temp_y * sn) / np.cos(by * np.pi / 180)
    dlat = temp_x * sn + temp_y * cs
    lon = dlon + bx
    lat = dlat + by
    return lon, lat


def create_tree(x, y, R=6371.0, leafsize=16):
    """Create a cKD tree object.

    Parameters
    ----------
    x,y: np.ndarray
        longitude and latitude of the grid location
    R: float
        The radius in kilometers of the planet.
    leafsize: int
        When to switch to brute force search.
    """
    if R:
        x, y, z = spherical2cartesian(lat=y, lon=x, R=R)
    else:
        z = xr.zeros_like(y)

    # Stack
    ridiculous_value = 777777
    if isinstance(x, xr.DataArray):
        x = x.fillna(ridiculous_value).data.ravel()
        y = y.fillna(ridiculous_value).data.ravel()
        z = z.fillna(ridiculous_value).data.ravel()
    elif isinstance(x, np.ndarray):
        x = x.ravel()
        np.nan_to_num(x.ravel(), nan=ridiculous_value, copy=False)
        y = y.ravel()
        np.nan_to_num(y.ravel(), nan=ridiculous_value, copy=False)
        z = z.ravel()
        np.nan_to_num(z.ravel(), nan=ridiculous_value, copy=False)

    # Construct KD-tree
    tree = spatial.cKDTree(np.vstack((x, y, z)).T, leafsize=leafsize)

    return tree


def NoneIn(lst):
    """See if there is a None in the iterable object. Return a Boolean."""
    ans = False
    for i in lst:
        if i is None:
            ans = True
            break
    return ans


@compileable
def find_ind(array, value, peri=None, ascending=1, above=True):
    """Find the index of the nearest value to the given value.

    Parameters
    ----------
    array: numpy.ndarray
        1D numpy array to search index from
    value: number
        The value to find nearest neighbor with
    peri: number, optional
        The periodicity of the array.
        For example, 360 for longitude.
    ascending: int, default 1
        Whether the array is in ascending order.
        1 for ascending order, -1 for descending order.
    above: boolean, default True
        If True, return the index of the largest item in
        array smaller than value.
        Otherwise, return the closest value.
    """
    array = np.asarray(array)
    if peri is None:
        index = np.argmin(np.abs(array - value))
    else:
        index = np.argmin(np.abs(to_180((array - value), peri=peri)))
    if above and array[index] > value and peri is None:
        index -= ascending * 1
    if index < 0 or index >= len(array):
        raise ValueError("Value out of bound.")
    index = int(index)
    return index, array[index]


deg2m = 6271e3 * np.pi / 180


def find_ind_h(lons, lats, tree, h_shape):
    """Use ckd tree to find the horizontal indexes,."""
    x, y, z = spherical2cartesian(lats, lons)
    _, index1d = tree.query(np.array([x, y, z]).T)
    if len(h_shape) == 3:
        face, iy, ix = np.unravel_index((index1d), h_shape)
    elif len(h_shape) == 2:
        face = None
        iy, ix = np.unravel_index((index1d), h_shape)
    return face, iy, ix


@compileable
def find_rel(
    value, array, darray=None, ascending=1, above=True, peri=None, dx_right=True
):
    """Find the rel-coords of the 1D coords.

    The backend for all find_rel functions

    Parameters
    ----------
    value: numpy.ndarray
        1D array for the value to find rel-coords.
    array: numpy.ndarray
        The array of potential reference levels.
    darray: numpy.ndarray, optional
        The distances between reference levels.
    peri: number, optional
        The periodicity of the array.
        For example, 360 for longitude.
    ascending: int, default 1
        Whether the array is in ascending order.
        1 for ascending order, -1 for descending order.
    above: boolean, default True
        If True, return the index of the largest item in
        array smaller than value.
        Otherwise, return the closest value.
    dx_right: boolean, default True
        If True, darray[i] = abs(array[i+1] - array[i])

    Returns
    -------
    ix: numpy.ndarray
        Indexes of the reference level
    rx: numpy.ndarray
        Non-dimensional distance to the reference level
    dx: numpy.ndarray
        distance between the reference t level and the next one.
    bx: numpy.ndarray
        Value of the reference level
    """
    if darray is None:
        darray = np.abs(array[1:] - array[:-1])
        darray = np.append(darray, darray[-1])
        dx_right = True
    ix_arr = np.zeros(len(value), dtype="int")
    rx_arr = np.ones_like(value) * 0.0
    dx_arr = np.ones_like(value) * 0.0
    bx_arr = np.ones_like(value) * 0.0

    dx_offset = int(not dx_right) + int(ascending > 0) - 1
    for i, x in enumerate(value):
        ix, bx = find_ind(array, x, ascending=ascending, above=above, peri=peri)
        if peri is None:
            dx = x - bx
        else:
            dx = to_180(x - bx, peri=peri)

        if not above or peri is not None:
            # peri not None will effectively overwrite above
            dx_offset = int(not dx_right) + int(ascending * dx > 0) - 1
        index = ix + dx_offset

        ix_arr[i] = ix
        rx_arr[i] = dx / darray[index]
        dx_arr[i] = darray[index]
        bx_arr[i] = bx
    return ix_arr, rx_arr, dx_arr, bx_arr


# Here are a few partial functions for find_rel
@compileable
def find_rel_nearest(value, ts):
    """Find the rel-coords based on the find_ind_nearest method."""
    return find_rel(value, ts, above=False)


@compileable
def find_rel_periodic(value, ts, peri):
    """Find the rel-coords based on the find_ind_periodic method."""
    return find_rel(value, ts, peri=peri)


@compileable
def find_rel_z(depth, some_z, some_dz=None, dz_above_z=True):
    """Find the rel-coords of the vertical coords.

    Parameters
    ----------
    depth: numpy.ndarray
        1D array for the depth of interest in meters.
        More negative means deeper.
    some_z: numpy.ndarray
        The depth of reference depth.
    some_dz: numpy.ndarray or None
        dz_i = abs(z_{i+1}- z_i)
    dz_above_z: Boolean
        Whether the dz as the distance between the depth level and
        a shallower one(True) or a deeper one(False)

    Returns
    -------
    iz: numpy.ndarray
        Indexes of the reference z level
    rz: numpy.ndarray
        Non-dimensional distance to the reference z level
    dz: numpy.ndarray
        distance between the reference z level and the next one.
    """
    return find_rel(
        depth, some_z, darray=some_dz, ascending=-1, dx_right=(not dz_above_z)
    )


@compileable
def find_rel_time(time, ts):
    """Find the rel-coords of the temporal coords.

    Parameters
    ----------
    time: numpy.ndarray
        1D array for the time since 1970-01-01 in seconds.
    ts: numpy.ndarray
        The time of model time steps also in seconds.

    Returns
    -------
    it: numpy.ndarray
        Indexes of the reference t level
    rt: numpy.ndarray
        Non-dimensional distance to the reference t level
    dt: numpy.ndarray
        distance between the reference t level and the next one.
    """
    return find_rel(time, ts)


def _read_h(some_x, some_y, some_dx, some_dy, CS, SN, ind):
    """Read the grid coords at given index.

    Parameters
    ----------
    some_x: numpy.ndarray
        array of longitude, could be XC or XG
    some_y: numpy.ndarray
        array of latitude, could be YC or YG
    some_dx: numpy.ndarray or None
        array of distances between grid in the longitudinal direction.
    some_dy: numpy.ndarray or None
        array of distances between grid in the latitudinal direction.
    CS: numpy.ndarray or None
        array of the cosine of the angle between grid and meridian.
    SN: numpy.ndarray or None
        array of the sine of the angle between grid and meridian.
    ind: tuple
        indexes to read the grid data from.
    """
    bx = some_x[ind]
    by = some_y[ind]
    if some_dx is not None and some_dy is not None:
        dx = some_dx[ind]
        dy = some_dy[ind]
    else:
        dx = None
        dy = None

    if CS is not None and SN is not None:
        cs = CS[ind]
        sn = SN[ind]
    else:
        cs = None
        sn = None
    return cs, sn, dx, dy, bx, by


def find_px_py(XG, YG, tp, ind, cuvwg="G"):
    """Find the nearest 4 corner points.

    This is used in oceanparcel interpolation scheme.
    """
    N = len(ind[0])
    ind1 = tuple(i for i in tp.ind_tend_vec(ind, np.ones(N) * 3, cuvwg=cuvwg))
    ind2 = tuple(i for i in tp.ind_tend_vec(ind1, np.zeros(N), cuvwg=cuvwg))
    ind3 = tuple(i for i in tp.ind_tend_vec(ind, np.zeros(N), cuvwg=cuvwg))

    x0 = XG[ind]
    x1 = XG[ind1]
    x2 = XG[ind2]
    x3 = XG[ind3]

    y0 = YG[ind]
    y1 = YG[ind1]
    y2 = YG[ind2]
    y3 = YG[ind3]

    px = np.vstack([x0, x1, x2, x3]).astype("float64")
    py = np.vstack([y0, y1, y2, y3]).astype("float64")

    return px, py


@compileable
def find_rx_ry_naive(x, y, bx, by, cs, sn, dx, dy):
    """Find the non-dimensional coords using the local cartesian scheme."""
    dlon = to_180(x - bx)
    dlat = to_180(y - by)
    rx = (dlon * np.cos(by * np.pi / 180) * cs + dlat * sn) * deg2m / dx
    ry = (dlat * cs - dlon * sn * np.cos(by * np.pi / 180)) * deg2m / dy
    return rx, ry


@compileable
def find_rx_ry_oceanparcel(x, y, px, py):
    """Find the non-dimensional horizontal distance.

    This is done using the oceanparcel scheme.
    """
    rx = np.ones_like(x) * 0.0
    ry = np.ones_like(y) * 0.0
    x0 = px[0]

    x = to_180(x - x0)
    px = to_180(px - x0)

    invA = np.array(
        [
            [1.0, 0.0, 0.0, 0.0],
            [-1.0, 1.0, 0.0, 0.0],
            [-1.0, 0.0, 0.0, 1.0],
            [1.0, -1.0, 1.0, -1.0],
        ]
    )
    a = np.dot(invA, px)
    b = np.dot(invA, py)

    aa = a[3] * b[2] - a[2] * b[3]
    bb = a[3] * b[0] - a[0] * b[3] + a[1] * b[2] - a[2] * b[1] + x * b[3] - y * a[3]
    cc = a[1] * b[0] - a[0] * b[1] + x * b[1] - y * a[1]

    det2 = bb * bb - 4 * aa * cc

    order1 = np.abs(aa) < 1e-12
    order2 = np.logical_and(~order1, det2 >= 0)
    ry = -(cc / bb)  # if it is supposed to be nan, just try linear solve.
    ry[order2] = ((-bb + np.sqrt(det2)) / (2 * aa))[order2]

    rot_rectilinear = np.abs(a[1] + a[3] * ry) < 1e-12
    rx[rot_rectilinear] = (
        (y - py[0]) / (py[1] - py[0]) + (y - py[3]) / (py[2] - py[3])
    )[rot_rectilinear] * 0.5
    rx[~rot_rectilinear] = ((x - a[0] - a[2] * ry) / (a[1] + a[3] * ry))[
        ~rot_rectilinear
    ]

    return rx - 1 / 2, ry - 1 / 2


def find_rel_h_naive(lon, lat, some_x, some_y, some_dx, some_dy, CS, SN, tree):
    """Find the rel-coords in the horizontal.

    very similar to find_rel_time/v
    rx,ry,dx,dy are defined the same way
    for example
    rx = "how much to the right of the node"/"size of the cell in left-right direction"
    dx = "size of the cell in left-right direction".

    cs,sn is just the cos and sin of the grid orientation.
    It will come in handy when we transfer vectors.
    """
    if NoneIn([lon, lat, some_x, some_y, some_dx, some_dy, CS, SN, tree]):
        raise ValueError("Some of the required variables are missing")
    h_shape = some_x.shape
    face, iy, ix = find_ind_h(lon, lat, tree, h_shape)
    if face is not None:
        ind = (face, iy, ix)
    else:
        ind = (iy, ix)
    cs, sn, dx, dy, bx, by = _read_h(some_x, some_y, some_dx, some_dy, CS, SN, ind)
    rx, ry = find_rx_ry_naive(lon, lat, bx, by, cs, sn, dx, dy)
    return face, iy, ix, rx, ry, cs, sn, dx, dy, bx, by


def find_rel_h_rectilinear(x, y, lon, lat):
    """Find the rel-coords using the rectilinear scheme."""
    ix, rx, dx, bx = find_rel_periodic(x, lon, 360.0)
    iy, ry, dy, by = find_rel_periodic(y, lat, 360.0)
    dx = np.cos(by * np.pi / 180) * deg2m * dx
    dy = deg2m * dy
    face = None
    cs = np.ones_like(x)
    sn = np.zeros_like(x)
    return face, iy, ix, rx, ry, cs, sn, dx, dy, bx, by


def find_rel_h_oceanparcel(
    x, y, some_x, some_y, some_dx, some_dy, CS, SN, XG, YG, tree, tp
):
    """Find the rel-coords using the rectilinear scheme."""
    if NoneIn([x, y, some_x, some_y, XG, YG, tree]):
        raise ValueError("Some of the required variables are missing")
    h_shape = some_x.shape
    face, iy, ix = find_ind_h(x, y, tree, h_shape)
    if face is not None:
        ind = (face, iy, ix)
    else:
        ind = (iy, ix)
    cs, sn, dx, dy, bx, by = _read_h(some_x, some_y, some_dx, some_dy, CS, SN, ind)
    px, py = find_px_py(XG, YG, tp, ind)
    rx, ry = find_rx_ry_oceanparcel(x, y, px, py)
    return face, iy, ix, rx, ry, cs, sn, dx, dy, bx, by


def weight_f_node(rx, ry):
    """Assign weights to four corners.

    assign weight based on the non-dimensional
    coords to the four corner points.
    """
    return np.vstack(
        [
            (0.5 - rx) * (0.5 - ry),
            (0.5 + rx) * (0.5 - ry),
            (0.5 + rx) * (0.5 + ry),
            (0.5 - rx) * (0.5 + ry),
        ]
    ).T


def find_cs_sn(thetaA, phiA, thetaB, phiB):
    """Find a spherical angle OAB.

    theta is the angle
    between the meridian crossing point A
    and the geodesic connecting A and B.

    this function return cos and sin of theta
    """
    # O being north pole
    AO = np.pi / 2 - thetaA
    BO = np.pi / 2 - thetaB
    dphi = phiB - phiA
    # Spherical law of cosine on AOB
    cos_AB = np.cos(BO) * np.cos(AO) + np.sin(BO) * np.sin(AO) * np.cos(dphi)
    sin_AB = np.sqrt(1 - cos_AB**2)
    # spherical law of sine on triangle AOB
    SN = np.sin(BO) * np.sin(dphi) / sin_AB
    CS = np.sign(thetaB - thetaA) * np.sqrt(1 - SN**2)
    return CS, SN


def missing_cs_sn(ds, return_xr=False):
    """Fill in the CS,SN of a dataset."""
    xc = np.deg2rad(np.array(ds.XC))
    yc = np.deg2rad(np.array(ds.YC))
    cs = np.zeros_like(xc)
    sn = np.zeros_like(xc)
    cs[0], sn[0] = find_cs_sn(yc[0], xc[0], yc[1], xc[1])
    cs[-1], sn[-1] = find_cs_sn(yc[-2], xc[-2], yc[-1], xc[-1])
    cs[1:-1], sn[1:-1] = find_cs_sn(yc[:-2], xc[:-2], yc[2:], xc[2:])
    if return_xr:
        cs = xr.DataArray(cs, dims=ds["XC"].dims)
        sn = xr.DataArray(sn, dims=ds["XC"].dims)

        return cs, sn
    else:
        return cs, sn


def convert_time(time):
    """Convert time into seconds after 1970-01-01.

    time needs to be a string or a np.datetime64 object.
    """
    t0 = np.datetime64("1970-01-01")
    one_sec = np.timedelta64(1, "s")
    if isinstance(time, str):
        dt = np.datetime64(time) - t0
        return dt / one_sec
    elif isinstance(time, np.datetime64):
        return (time - t0) / one_sec
    else:
        raise ValueError(f"Unsupported time format {type(time)}")


def easy_3d_cube(lon, lat, dep, tim, print_total_number=False):
    """Create 4D coords for initializing Position/Particle."""
    east, west, Nlon = lon
    south, north, Nlat = lat
    shallow, deep, Ndep = dep
    t_in_sec = convert_time(tim)

    x = np.linspace(east, west, Nlon)
    y = np.linspace(south, north, Nlat)
    x, y = np.meshgrid(x, y)
    x = x.ravel()
    y = y.ravel()
    levels = np.linspace(shallow, deep, Ndep)
    x, z = np.meshgrid(x, levels)
    y, z = np.meshgrid(y, levels)
    x = x.ravel()
    y = y.ravel()
    z = z.ravel()

    t = np.ones_like(x) * t_in_sec
    if print_total_number:
        print(f"A total of {len(x)} positions are defined.")
    return x, y, z, t


@compileable
def pointinpolygon(x, y, poly):
    n = len(poly)
    inside = False
    p2x = 0.0
    p2y = 0.0
    xints = 0.0
    p1x, p1y = poly[0]
    for i in numba.prange(n + 1):
        p2x, p2y = poly[i % n]
        if y > min(p1y, p2y):
            if y <= max(p1y, p2y):
                if x <= max(p1x, p2x):
                    if p1y != p2y:
                        xints = (y - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
                    if p1x == p2x or x <= xints:
                        inside = not inside
        p1x, p1y = p2x, p2y

    return inside


@compileable_parallel
def parallelpointinpolygon(xs, ys, poly):
    D = np.empty(len(xs), dtype=numba.boolean)
    for i in numba.prange(0, len(D)):
        D[i] = pointinpolygon(xs[i], ys[i], poly)
    return D


# functions used in Lagrangian models to avoid cyclic import
@compileable
def _increment(t, u, du):
    """Find how far it will travel in duration t.

    For a one dimensional particle with speed u and speed derivative du,
    find how far it will travel in duration t.

    Parameters
    ----------
    t: float, numpy.ndarray
        The time duration
    u: float, numpy.ndarray
        The velocity defined at the starting point.
    du: float, numpy.ndarray
        The velocity gradient. Assumed to be constant.
    """
    incr = u / du * (np.exp(du * t) - 1)
    no_gradient = np.abs(du) < 1e-12
    incr[no_gradient] = (u * t)[no_gradient]
    return incr


def _stationary(t, u, du, x0):
    """Find the final position after time t.

    For a one dimensional Particle with speed u and speed derivative du
    starting at x0, find the final position after time t.
    "Stationary" means that we are assuming there is no time dependency.

    Parameters
    ----------
    t: float, numpy.ndarray
        The time duration
    u: float, numpy.ndarray
        The velocity defined at the starting point.
    du: float, numpy.ndarray
        The velocity gradient. Assumed to be constant.
    x0: float, numpy.ndarray
        The starting position.
    """
    incr = _increment(t, u, du)
    return incr + x0


@compileable
def _stationary_time(u, du, x0):
    """Find the amount of time to leave the cell.

    Find the amount of time it needs for a Particle to hit x = -0.5 and 0.5.
    The time could be negative.

    Parameters
    ----------
    u: numpy.ndarray
        The velocity defined at the starting point.
    du: numpy.ndarray
        The velocity gradient. Assumed to be constant.
    x0: numpy.ndarray
        The starting position.

    Returns
    -------
    tl: numpy.ndarray
        The time it takes to hit -0.5.
    tr: numpy.ndarray
        The time it takes to hit 0.5
    """
    tl = np.log(1 - du / u * (0.5 + x0)) / du
    tr = np.log(1 + du / u * (0.5 - x0)) / du
    no_gradient = np.abs(du) < 1e-12
    tl[no_gradient] = (-x0[no_gradient] - 0.5) / u[no_gradient]
    tr[no_gradient] = (0.5 - x0[no_gradient]) / u[no_gradient]
    return tl, tr


@compileable
def _uleftright_from_udu(u, du, x0):
    """Calculate the velocity at -0.5 and 0.5."""
    u_left = u - (x0 + 0.5) * du
    u_right = u + (0.5 - x0) * du
    return u_left, u_right


def _time2wall(pos_list, u_list, du_list, tf):
    """Apply stationary_time three times for all three dimensions."""
    ts = []
    for i in range(3):
        tl, tr = _stationary_time(u_list[i], du_list[i], pos_list[i])
        ul, ur = _uleftright_from_udu(u_list[i], du_list[i], pos_list[i])
        sign = np.sign(tf)
        cannot_left = -ul * sign <= 1e-12  # aroung 30000 years
        tl[cannot_left] = -sign[cannot_left]
        cannot_right = ur * sign <= 1e-12
        tr[cannot_right] = -sign[cannot_right]
        ts.append(tl)
        ts.append(tr)
    return ts


def _which_early(tf, ts):
    """Find out which event happens first.

    We are trying to integrate the Particle to time tf.
    The first event is either:
    1. tf is reached before reaching a wall
    2. ts[i] is reached, and a Particle hit a wall. ts[i]*tf>0.

    Parameters
    ----------
    tf: float, numpy.ndarray
        The final time
    ts: list
        The list of events calculated using _time2wall
    """
    ts.append(np.ones(len(ts[0])) * tf)  # float or array both ok
    t_directed = np.array(ts) * np.sign(tf)
    t_directed[np.isnan(t_directed)] = np.inf
    t_directed[t_directed < 0] = np.inf
    tend = t_directed.argmin(axis=0)
    t_event = np.array([ts[te][i] for i, te in enumerate(tend)])
    return tend, t_event



import numpy as np
import pytest

import seaduck as sd
from seaduck import utils
from seaduck.lagrangian_budget import (
    find_ind_frac_tres,
    first_last_neither,
    flatten,
    ind_tend_uv,
    particle2xarray,
    redo_index,
    store_lists,
)


@pytest.fixture
def custom_pt():
    x = np.linspace(-50, -15, 5)
    y = np.ones_like(x) * 52.0
    z = np.ones_like(x) * (-9)
    t = np.ones_like(x)
    od = sd.OceData(utils.get_dataset("ecco"))
    pt = sd.Particle(
        x=x,
        y=y,
        z=z,
        t=t,
        data=od,
        uname="utrans",
        vname="vtrans",
        wname="wtrans",
        transport=True,
        save_raw=True,
    )
    pt.to_next_stop(t[0] + 1e7)
    return pt


@pytest.fixture
def curv_pt():
    od = sd.OceData(utils.get_dataset("curv"))
    curv_p = sd.Particle(
        y=np.array([70.5]),
        x=np.array([-14.0]),
        z=np.array([-10.0]),
        t=np.array([od.ts[0]]),
        data=od,
        uname="U",
        vname="V",
        wname="W",
        save_raw=True,
    )
    curv_p.to_next_stop(od.ts[0] + 1e4)
    return curv_p


@pytest.fixture
def region_info():
    GULF = np.array(
        [
            [-92.5, 25],
            [-75, 25],
            [-55, 40],
            [-72.5, 40],
        ]
    )
    LABR = np.array([[-72, 63.5], [-60, 50], [-49, 53], [-61, 66.5]])
    GDBK = np.array(
        [
            # GULF[2],
            GULF[3],
            LABR[1],
            LABR[2],
            [-42, 50],
            [-42, 40],
        ]
    )
    NACE = np.array([GDBK[-2], GDBK[-1], [-21, 47], [-21, 57]])
    EGRL = np.array([[-22.5, 72], [-44, 63], [-44, 57], [-22.5, 66]])

    return ["gulf", "labr", "gdbk", "nace", "egrl"], [GULF, LABR, GDBK, NACE, EGRL]


@pytest.mark.parametrize(
    "ind,exp",
    [
        ((0, 2, 45, 45), (0, 2, 45, 46)),
        ((1, 2, 45, 45), (1, 2, 46, 45)),
        ((1, 2, 89, 45), (0, 6, 44, 0)),
        ((0, 5, 45, 89), (1, 7, 0, 44)),
    ],
)
@pytest.mark.parametrize("od", ["ecco"], indirect=True)
def test_ind_tend_uv(ind, exp, od):
    tub = od
    tp = tub.tp
    ans = ind_tend_uv(ind, tp)
    assert exp == ans


def test_ind_tend_uv_error():
    ind = (4, 3, 2, 1)
    with pytest.raises(ValueError):
        ind_tend_uv(ind, None)


def test_redo_index(custom_pt):
    vf, vb, frac = redo_index(custom_pt)
    assert (frac <= 1).all()
    assert (frac >= 0).all()


def test_flatten_list():
    lst = [[0, 1], [1, 1, 1, 1]]
    new = flatten(lst)
    assert (new == np.array([0, 1, 1, 1, 1, 1])).all()
    assert isinstance(new, np.ndarray)


@pytest.mark.parametrize("od", ["ecco"], indirect=True)
def test_ind_frac_find(custom_pt, od):
    particle_datasets = particle2xarray(custom_pt)
    tub = od
    ind1, ind2, frac, tres, last, first = find_ind_frac_tres(particle_datasets, tub)
    assert ind1.shape[0] == 5
    assert (frac != 1).any()
    assert (tres >= 0).all()


@pytest.mark.parametrize("od", ["curv"], indirect=True)
def test_ind_frac_find_noface(curv_pt, od):
    particle_datasets = particle2xarray(curv_pt)
    tub = od
    ind1, ind2, frac, tres, last, first = find_ind_frac_tres(particle_datasets, tub)
    assert ind1.shape[0] == 4
    assert (frac != 1).any()
    assert (tres >= 0).all()


def test_store_lists(custom_pt):
    store_lists(custom_pt, "PleaseIgnore_dump.zarr")


def test_store_lists_with_region(custom_pt, region_info):
    pytest.importorskip("numba")
    region_names, region_polys = region_info
    store_lists(
        custom_pt,
        "PleaseIgnore_dump.zarr",
        region_names=region_names,
        region_polys=region_polys,
    )


def test_first_last_neither():
    shapes = np.random.randint(2, 5, 10)
    first, last, neither = first_last_neither(shapes)
    merged = np.sort(np.concatenate([first, last, neither]))
    sums = np.sum(shapes)
    assert np.allclose(np.arange(sums), merged)



import copy
import logging
from functools import cache
from itertools import combinations

import numpy as np

from seaduck.runtime_conf import compileable

# default kernel for interpolation.
DEFAULT_KERNEL = np.array(
    [[0, 0], [0, 1], [0, 2], [0, -1], [0, -2], [-1, 0], [-2, 0], [1, 0], [2, 0]]
)
DEFAULT_INHERITANCE = [
    [0, 1, 2, 3, 4, 5, 6, 7, 8],
    [0, 1, 2, 3, 5, 7, 8],
    [0, 1, 3, 5, 7],
    [0],
]

DEFAULT_KERNELS = [
    np.array([DEFAULT_KERNEL[i] for i in doll]) for doll in DEFAULT_INHERITANCE
]


# It just tell you what the kernels look like
def show_kernels(kernels=DEFAULT_KERNELS):
    """Plot a small scatter plot of the shape of a list of kernel.

    Parameters
    ----------
    kernels: list of numpy.ndarray
        Each of the element is a (n,2) shaped array,
        where n is the number of element in the kernel.
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError as exc:
        raise ImportError("maptlotlib.pyplot is needed to use this function.") from exc

    for i, k in enumerate(kernels):
        x, y = k.T
        plt.plot(x + 0.1 * i, y + 0.1 * i, "+")


def _translate_to_tendency(kernel):
    """Translate a movement to directions.

    A kernel looks like
    np.array([
    [x1,y1],
    [x2,y2],....
    ])
    where [xn,yn] represent a coordinate relative to [0,0]
    which is just the nearest neighbor.
    this function return how you could move from [0,0] to k = [x,y]
    If you need to go to [0,2], you need to move up twice.
    it will return [0,0] or more explicitly ['up','up']
    [0,0] will produce a empty array.

    Parameters
    ----------
    k: numpy.ndarray
        A (n,2)-shaped array, where n is the number of
        element in the kernel.
    """
    tend = []
    x, y = kernel
    if y > 0:
        for j in range(y):
            tend.append(0)  # up
    else:
        for j in range(-y):
            tend.append(1)  # down
    if x < 0:
        for j in range(-x):
            tend.append(2)  # left
    else:
        for j in range(x):
            tend.append(3)  # right
    return tend


def kernel_weight_x(kernel, kernel_type="interp", order=0):
    r"""Return the function that calculate the interpolation/derivative weight.

    input needs to be a cross-shaped (that's where x is coming from) Lagrangian kernel.

    If you don't want to know what is going on under the hood.
    it's totally fine.

    all of the following is a bit hard to understand.
    The k th (k>=0) derivative of the lagrangian polynomial is
    $$
    w_j= \frac{\Sigma_{i\neq j} Pi_{i<m-1-k} (x-x_i)}{\Pi_{i\neq j} (x_j - x_i)}
    $$
    for example: if the points are [-1,0,1] for point 0
    k = 0: w = (x-1)(x+1)/(0-1)(0+1)
    k = 1: w = [(x+1)+(x-1)]/(0-1)(0+1)

    for a cross shape kernel:
    f(rx,ry) = f_x(rx) + f_y(ry) - f(0,0)

    Parameters
    ----------
    kernel: np.ndarray
        2D array with shape (n,2), where n is the number of nodes.
        It has to be shaped like a cross
    kernel_type: str
        "interp" (default): Use both x y direction for interpolation,
        implies that order = 0
        "x": Use only x direction for interpolation/derivative
        "y": Use only y direction for interpolation/derivative
    order: int
        The order of derivatives. Zero for interpolation.

    Returns
    -------
    func(rx,ry): compilable function
        function to calculate the hotizontal interpolation/derivative
        weight
    """
    xs = np.array(list(set(kernel.T[0])), dtype=float)
    ys = np.array(list(set(kernel.T[1])), dtype=float)

    # if you the kernel is a line rather than a cross
    if len(xs) == 1:
        kernel_type = "y"
    elif len(ys) == 1:
        kernel_type = "x"

    x_poly = []
    y_poly = []
    if kernel_type == "interp":
        for ax in xs:
            x_poly.append(list(combinations([i for i in xs if i != ax], len(xs) - 1)))
        for ay in ys:
            y_poly.append(list(combinations([i for i in ys if i != ay], len(ys) - 1)))
    if kernel_type == "x":
        for ax in xs:
            x_poly.append(
                list(combinations([i for i in xs if i != ax], len(xs) - 1 - order))
            )
        y_poly = [[[]]]
    if kernel_type == "y":
        x_poly = [[[]]]
        for ay in ys:
            y_poly.append(
                list(combinations([i for i in ys if i != ay], len(ys) - 1 - order))
            )
    x_poly = np.array(x_poly, dtype=float)
    y_poly = np.array(y_poly, dtype=float)

    @compileable
    def the_interp_func(rx, ry):
        nonlocal kernel, xs, ys, x_poly, y_poly
        n = len(rx)
        m = len(kernel)
        weight = np.ones((n, m)) * 0.0
        for i, (x, y) in enumerate(kernel):
            if x != 0:
                ix = np.where(xs == x)[0][0]
                poly = x_poly[ix]
                for term in poly:
                    another = np.ones(n)
                    for other in term:
                        another *= rx - other
                    weight[:, i] += another
                    for other in xs:
                        if other != x:
                            weight[:, i] /= x - other
            if y != 0:
                iy = np.where(ys == y)[0][0]
                poly = y_poly[iy]
                for term in poly:
                    another = np.ones(n) * 1.0
                    for other in term:
                        another *= ry - other
                    weight[:, i] += another
                    for other in ys:
                        if other != y:
                            weight[:, i] /= y - other
            elif x**2 + y**2 == 0:
                xthing = np.zeros(n) * 0.0
                ix = np.where(xs == 0)[0][0]
                poly = x_poly[ix]
                for term in poly:
                    another = np.ones(n) * 1.0
                    for other in term:
                        another *= rx - other
                    xthing += another
                    for other in xs:
                        if other != x:
                            xthing /= x - other

                ything = np.zeros(n) * 0.0
                iy = np.where(ys == y)[0][0]
                poly = y_poly[iy]
                for term in poly:
                    another = np.ones(n)
                    for other in term:
                        another *= ry - other
                    ything += another
                    for other in ys:
                        if other != y:
                            ything /= y - other
                weight[:, i] = xthing + ything - 1
        return weight

    @compileable
    def the_x_func(rx, ry):
        nonlocal kernel, xs, ys, x_poly, order
        n = len(rx)
        m = len(kernel)
        weight = np.ones((n, m)) * 0.0
        for i, (x, y) in enumerate(kernel):
            if y == 0:
                ix = np.where(xs == x)[0][0]
                poly = x_poly[ix]
                for term in poly:
                    another = np.ones(n) * 1.0
                    for other in term:
                        another *= rx - other

                    weight[:, i] += another
                for other in xs:
                    if other != x:
                        weight[:, i] /= x - other
        return weight

    @compileable
    def the_x_maxorder_func(rx, ry):
        nonlocal kernel, xs, ys, order
        n = len(rx)
        m = len(kernel)
        common = 1
        for i in range(1, order):
            common *= i
        weight = np.ones((n, m)) * float(common)
        for i, (x, y) in enumerate(kernel):
            if y == 0:
                for other in xs:
                    if other != x:
                        weight[:, i] /= x - other
            else:
                weight[:, i] = 0.0
        return weight

    @compileable
    def the_y_func(rx, ry):
        nonlocal kernel, xs, ys, y_poly, order
        n = len(rx)
        m = len(kernel)
        weight = np.ones((n, m)) * 0.0
        for i, (x, y) in enumerate(kernel):
            if x == 0:
                iy = np.where(ys == y)[0][0]
                poly = y_poly[iy]
                for term in poly:
                    another = np.ones(n) * 1.0
                    for other in term:
                        another *= ry - other

                    weight[:, i] += another
                for other in ys:
                    if other != y:
                        weight[:, i] /= y - other
        return weight

    @compileable
    def the_y_maxorder_func(rx, ry):
        nonlocal kernel, xs, ys, order
        n = len(rx)
        m = len(kernel)
        common = 1
        for i in range(1, order):
            common *= i
        weight = np.ones((n, m)) * float(common)
        for i, (x, y) in enumerate(kernel):
            if x == 0:
                for other in ys:
                    if other != y:
                        weight[:, i] /= y - other
            else:
                weight[:, i] = 0.0
        return weight

    if kernel_type == "interp":
        return the_interp_func
    if kernel_type == "x":
        max_order = len(xs) - 1
        if order < max_order:
            return the_x_func
        elif order == max_order:
            return the_x_maxorder_func
        else:
            raise ValueError("Kernel is too small for this derivative")
    if kernel_type == "y":
        max_order = len(ys) - 1
        if order < max_order:
            return the_y_func
        elif order == max_order:
            return the_y_maxorder_func
        else:
            raise ValueError("Kernel is too small for this derivative")


# we can define the default interpolation functions here,
# so if we are using it over and over, we don't have to compile it.
# and it really takes a lot of time to compile.


def kernel_weight_s(kernel, xorder=0, yorder=0):
    """Return the function that calculate the interpolation/derivative weight.

    input needs to be a rectangle-shaped (that's where x is coming from)
    Lagrangian kernel.

    Parameters
    ----------
    kernel: np.ndarray
        2D array with shape (n,2), where n is the number of nodes.
        It has to be shaped like a rectangle
    xorder: int
        The order of derivatives in the x direction.
        Zero for interpolation.
    yorder: int
        The order of derivatives in the y direction.
        Zero for interpolation.

    Returns
    -------
    func(rx,ry): compilable function
        function to calculate the hotizontal interpolation/derivative
        weight
    """
    xs = np.array(list(set(kernel.T[0])), dtype=float)
    ys = np.array(list(set(kernel.T[1])), dtype=float)
    xmaxorder = False
    ymaxorder = False
    if xorder < len(xs) - 1:
        pass
    elif xorder == len(xs) - 1:
        xmaxorder = True
    else:
        raise ValueError("Kernel is too small for this derivative")

    if yorder < len(ys) - 1:
        pass
    elif yorder == len(ys) - 1:
        ymaxorder = True
    else:
        raise ValueError("Kernel is too small for this derivative")

    x_poly = []
    y_poly = []
    for ax in xs:
        x_poly.append(
            list(combinations([i for i in xs if i != ax], len(xs) - 1 - xorder))
        )
    for ay in ys:
        y_poly.append(
            list(combinations([i for i in ys if i != ay], len(ys) - 1 - yorder))
        )
    x_poly = np.array(x_poly, dtype=float)
    y_poly = np.array(y_poly, dtype=float)

    @compileable
    def the_square_func(rx, ry):
        nonlocal kernel, xs, ys, y_poly, x_poly, xorder, yorder, xmaxorder, ymaxorder
        n = len(rx)
        num_node_x = len(xs)
        num_node_y = len(ys)
        m = len(kernel)
        yweight = np.ones((n, num_node_y))
        xweight = np.ones((n, num_node_x))
        weight = np.ones((n, m)) * 0.0

        if ymaxorder:
            common = 1
            for i in range(1, yorder):
                common *= i
            yweight *= float(common)
        else:
            yweight *= 0.0
        for i, y in enumerate(ys):
            if not ymaxorder:
                iy = np.where(ys == y)[0][0]
                poly = y_poly[iy]
                for term in poly:
                    another = np.ones(n) * 1.0
                    for other in term:
                        another *= ry - other

                    yweight[:, i] += another
            for other in ys:
                if other != y:
                    yweight[:, i] /= y - other

        if xmaxorder:
            common = 1
            for i in range(1, xorder):
                common *= i
            xweight *= float(common)
        else:
            xweight *= 0.0
        for i, x in enumerate(xs):
            if not xmaxorder:
                ix = np.where(xs == x)[0][0]
                poly = x_poly[ix]
                for term in poly:
                    another = np.ones(n) * 1.0
                    for other in term:
                        another *= rx - other

                    xweight[:, i] += another
            for other in xs:
                if other != x:
                    xweight[:, i] /= x - other

        for i, (x, y) in enumerate(kernel):
            iy = np.where(ys == y)[0][0]
            ix = np.where(xs == x)[0][0]
            weight[:, i] = yweight[:, iy] * xweight[:, ix]

        return weight

    return the_square_func


def kernel_weight(kernel, kernel_type="interp", order=0):
    """Return a function that compute weights.

    A wrapper around kernel_weight_x and kernel_weight_s.
    Return the function that calculate the interpolation/derivative weight
    of a  Lagrangian kernel.

    Parameters
    ----------
    kernel: np.ndarray
        2D array with shape (n,2), where n is the number of nodes.
        It need to either shape like a rectangle or a cross
    kernel_type: str
        "interp" (default): Use both x y direction for interpolation,
        implies that order = 0
        "dx": Use only x direction for interpolation/derivative
        "dy": Use only y direction for interpolation/derivative
    order: int
        The order of derivatives. Zero for interpolation.

    Returns
    -------
    func(rx,ry): compilable function
        function to calculate the hotizontal interpolation/derivative
        weight
    """
    num_node_x = len(set(kernel[:, 0]))
    num_node_y = len(set(kernel[:, 1]))
    if len(kernel) == num_node_x + num_node_y - 1:
        if "d" in kernel_type:
            kernel_type = kernel_type[1:]
        return kernel_weight_x(kernel, kernel_type=kernel_type, order=order)
    elif len(kernel) == num_node_x * num_node_y:
        # num_node_x*num_node_y == num_node_x+num_node_y-1
        # only when num_node_x==1 or num_node_y ==1
        if kernel_type == "interp":
            return kernel_weight_s(kernel, xorder=0, yorder=0)
        elif kernel_type == "dx":
            return kernel_weight_s(kernel, xorder=order, yorder=0)
        elif kernel_type == "dy":
            return kernel_weight_s(kernel, xorder=0, yorder=order)
        else:
            raise ValueError(f"kernel_type = {kernel_type} not supported")
    else:
        raise NotImplementedError("The shape of the kernel is neither cross or square")


default_interp_funcs = [kernel_weight_x(a_kernel) for a_kernel in DEFAULT_KERNELS]


def find_which_points_for_each_kernel(masked, inheritance="default"):
    """Find which kernel to use at each point.

    masked is going to be a n*m array,
    where n is the number of points of interest.
    m is the size of the largest kernel.

    inheritance defines the shape of smaller kernels.
    say
    inheritance = [
    [0,1,2,3,4],
    [0,1],
    [0]
    ]
    it means that the largest kernel have all 5 nodes
    the second kernel only contain the first and second node,
    and the last one only have the nearest neighbor.

    if a row of matrix looks like [1,1,1,1,1],
    the index of the row will be in the first element(list) of the return
    variable.

    if a row of matrix looks like [1,1,1,1,0],
    although it fits both 2nd and 3rd kernel, 2nd has priority,
    so the index will be in the 2nd element of the return pk.

    if a row looks like [0,0,0,0,0],
    none of the kernel can fit it, so the index will not be in the
    return
    """
    if inheritance == "default":
        inheritance = DEFAULT_INHERITANCE
    already_wet = []
    for i, doll in enumerate(inheritance):
        wet_1d = masked[:, np.array(doll)].all(axis=1)
        already_wet.append(np.where(wet_1d)[0])
    point_for_each_kernel = [list(already_wet[0])]
    for i in range(1, len(inheritance)):
        point_for_each_kernel.append(
            list(np.setdiff1d(already_wet[i], already_wet[i - 1]))
        )
    return point_for_each_kernel


def get_weight_cascade(
    rx,
    ry,
    pk,
    kernel_large=DEFAULT_KERNEL,
    inheritance=DEFAULT_INHERITANCE,
    funcs=default_interp_funcs,
):
    """Compute the weight.

    apply the corresponding functions that was figured out in
    find_which_points_for_each_kernel

    Parameters
    ----------
    rx,ry: numpy.ndarray
        1D array with length N of non-dimensional relative horizontal
        position to the nearest node
    kernel_large: numpy.ndarray
        A numpy kernel of shape (M,2) that contains all the kernels needed.
    inheritance: list of list(s)
        The inheritance sequence when some of the node is masked.
    funcs: list of compileable functions
        The weight function of each kernel in the inheritance sequence.

    Returns
    -------
    + weight: numpy.ndarray
        The horizontal weight of interpolation/derivative for the points
        with shape (N,M)

    """
    weight = np.zeros((len(rx), len(kernel_large)))
    weight[:, 0] = np.nan
    for i in range(len(pk)):
        if len(pk[i]) == 0:
            continue
        sub_rx = rx[pk[i]]
        sub_ry = ry[pk[i]]
        #     slim_weight = interp_func[i](sub_rx,sub_ry)
        sub_weight = np.zeros((len(pk[i]), len(kernel_large)))
        sub_weight[:, np.array(inheritance[i])] = funcs[i](sub_rx, sub_ry)
        weight[pk[i]] = sub_weight
    return weight


def _find_pk_4d(masked, inheritance=DEFAULT_INHERITANCE):
    """Find the masking condition for 4D space time.

    See find_which_points_for_each_kernel
    """
    maskedT = masked.T
    ind_shape = maskedT.shape
    pk4d = []
    for i in range(ind_shape[0]):
        z = []
        for j in range(ind_shape[1]):
            z.append(find_which_points_for_each_kernel(maskedT[i, j].T, inheritance))
        pk4d.append(z)
    return pk4d


def _kash(kernel):  # hash kernel
    """Hash a horizontal kernel.

    Return the hash value.

    Parameters
    ----------
    + kernel: numpy.ndarray
        A horizontal kernel
    """
    return hash(tuple((i, j) for (i, j) in kernel))


@cache
def _get_func_from_hashable(
    kernel_tuple, kernel_shape, hkernel="interp", h_order=0, **kwargs
):
    kernel = np.array(kernel_tuple).reshape(kernel_shape)
    return kernel_weight(kernel, kernel_type=hkernel, order=h_order)


def get_func(kernel, **kwargs):
    """Return functions that compute weights.

    Similar to the kernel_weight function,
    the only difference is that this function can
    read existing functions that is cached.
    See _get_func_from_hashable

    See Also
    --------
    kernel_weight: the un-hashed version of this function.
    """
    return _get_func_from_hashable(tuple(kernel.ravel()), kernel.shape, **kwargs)


def _auto_inheritance(kernel, hkernel="interp"):
    """Find a natural inheritance pattern given one horizontal kernel."""
    if hkernel == "interp":
        doll = [list(range(len(kernel)))]
    elif hkernel == "dx":
        doll = [[i for i in range(len(kernel)) if kernel[i][1] == 0]]
    elif hkernel == "dy":
        doll = [[i for i in range(len(kernel)) if kernel[i][0] == 0]]
    doll[0] = sorted(
        doll[0], key=lambda i: max(abs(kernel[i] + np.array([0.01, 0.00618])))
    )
    last = doll[-1]
    lask = np.array([kernel[i] for i in last])
    dist = round(max(np.max(abs(lask), axis=1)))
    for radius in range(dist - 1, -1, -1):
        new = [i for i in last if max(abs(kernel[i])) <= radius]
        if new != last:
            last = new
            doll.append(last)
    return doll


class KnW:
    """Kernel object.

    A class that describes anything about the
    interpolation/derivative kernel to be used.

    Parameters
    ----------
    kernel: numpy.ndarray
        The largest horizontal kernel to be used
    inheritance: list
        The inheritance sequence of the kernels
    hkernel: str
        What to do in the horizontal direction
        'interp', 'dx', or 'dy'?
    tkernel: str
        What kind of operation to do in the temporal dimension:
        'linear', 'nearest' interpolation, or 'dt'
    vkernel: str
        What kind of operation to do in the vertical:
        'linear', 'nearest' interpolation, or 'dz'
    h_order: int
        How many derivative to take in the horizontal direction.
        Zero for pure interpolation
    ignore_mask: bool
        Whether to diregard the masking of the dataset.
        You can select True if there is no
        inheritance, or if performance is a big concern.
    """

    def __init__(
        self,
        kernel=DEFAULT_KERNEL,
        inheritance="auto",  # None, or list of lists
        hkernel="interp",  # dx,dy
        vkernel="nearest",  # linear,dz
        tkernel="nearest",  # linear,dt
        h_order=0,  # depend on hkernel type
        ignore_mask=False,
    ):
        kernel_sort = np.abs(kernel + np.array([0.01, 0.00618])).sum(axis=1).argsort()
        # Avoid points having same distance
        kernel_sort_inv = kernel_sort.argsort()

        if (inheritance is not None) and (ignore_mask):
            logging.info(
                "Overwriting the inheritance object to None,"
                " because we ignore masking"
            )

        if inheritance == "auto":
            inheritance = _auto_inheritance(kernel, hkernel=hkernel)
        elif inheritance is None:  # does not apply cascade
            inheritance = [list(range(len(kernel)))]
        elif isinstance(inheritance, list):
            pass
        else:
            raise ValueError("Unknown type of inherirance")

        self.kernel = kernel[kernel_sort]
        self.inheritance = [
            sorted([kernel_sort_inv[i] for i in heir]) for heir in inheritance
        ]
        self.hkernel = hkernel
        self.vkernel = vkernel
        self.tkernel = tkernel
        self.h_order = h_order
        self.ignore_mask = ignore_mask

        self.kernels = [
            np.array([self.kernel[i] for i in doll]) for doll in self.inheritance
        ]
        self.hfuncs = [
            get_func(kernel=a_kernel, hkernel=self.hkernel, h_order=self.h_order)
            for a_kernel in self.kernels
        ]

    def same_hsize(self, other):
        """Return True if 2 KnW object has the same horizontal size."""
        type_same = isinstance(other, type(self))
        if not type_same:
            raise TypeError("the argument is not a KnW object")
        try:
            return np.allclose(self.kernel, other.kernel)
        except (ValueError, AttributeError):
            return False

    def same_size(self, other):
        """Return True if 2 KnW object has the same 4D size."""
        only_size = {"dz": 2, "linear": 2, "dt": 2, "nearest": 1}
        hsize_same = self.same_hsize(other)
        vsize_same = only_size[self.vkernel] == only_size[other.vkernel]
        tsize_same = only_size[self.tkernel] == only_size[other.tkernel]
        return hsize_same and vsize_same and tsize_same

    def __eq__(self, other):
        type_same = isinstance(other, type(self))
        if not type_same:
            return False
        shpe_same = self.same_hsize(other) and self.inheritance == other.inheritance
        diff_same = (
            (self.hkernel == other.hkernel)
            and (self.vkernel == other.vkernel)
            and (self.tkernel == other.tkernel)
        )
        return type_same and shpe_same and diff_same

    def __hash__(self):
        return hash(
            (
                _kash(self.kernel),
                tuple(tuple(i for i in heir) for heir in self.inheritance),
                self.ignore_mask,
                self.h_order,
                self.hkernel,
                self.vkernel,
                self.tkernel,
            )
        )

    def size_hash(self):
        """Produce a hash value based on the 4D size of the KnW object."""
        only_size = {"dz": 2, "linear": 2, "dt": 2, "nearest": 1}
        return hash(
            (_kash(self.kernel), only_size[self.vkernel], only_size[self.tkernel])
        )

    def get_weight(
        self,
        rx,
        ry,
        rz=0,
        rt=0,
        pk4d=None,  # All using the largest
        bottom_scheme="no flux",  # None
    ):
        """Return the weight of values given particle rel-coords.

        Parameters
        ----------
        rx,ry,rz,rt: numpy.ndarray
            1D array of non-dimensional particle positions
        pk4d: list
            A mapping on which points should use which kernel.
        bottom_scheme: str
            Whether to assume there is a ghost point with same value at
            the bottom boundary.
            Choose None for vertical flux, 'no flux' for most other cases.

        Returns
        -------
        weight: numpy.ndarray
            The weight of interpolation/derivative for the points
            with shape (N,M),
            M is the num of node in the largest kernel.
        """
        if self.vkernel in ["linear", "dz"]:
            nz = 2
        else:
            nz = 1
        if self.tkernel in ["linear", "dt"]:
            nt = 2
        else:
            nt = 1

        weight = np.zeros((len(rx), len(self.kernel), nz, nt))
        if isinstance(rz, (int, float)) and self.vkernel != "nearest":
            rz = np.array([rz for i in range(len(rx))])
        if isinstance(rt, (int, float)) and self.tkernel != "nearest":
            rt = np.array([rt for i in range(len(rx))])

        if self.tkernel == "linear":
            rpt = copy.deepcopy(rt)
            tweight = [
                (1 - rpt).reshape((len(rpt), 1, 1)),
                rpt.reshape((len(rpt), 1, 1)),
            ]
        elif self.tkernel == "dt":
            tweight = [-1, 1]
        elif self.tkernel == "nearest":
            tweight = [1]

        if self.vkernel == "linear":
            rpz = copy.deepcopy(rz)
            zweight = [(1 - rpz).reshape((len(rpz), 1)), rpz.reshape((len(rpz), 1))]
        elif self.vkernel == "dz":
            zweight = [-1, 1]
        elif self.vkernel == "nearest":
            zweight = [1]

        if pk4d is None:
            # pk4d = [
            #    [
            #        [list(range(len(rx)))]
            # # all points are in the same catagory
            #    for i in range(nz)]# Every layer is the same
            # for j in range(nt)]#
            for jt in range(nt):
                for jz in range(nz):
                    weight[:, self.inheritance[0], jz, jt] = self.hfuncs[0](rx, ry)
        else:
            if nt != len(pk4d) or nz != len(pk4d[0]):
                raise ValueError("The kernel and the input pk4d does not match")

            for jt in range(nt):
                for jz in range(nz):
                    weight[:, :, jz, jt] = get_weight_cascade(
                        rx,
                        ry,
                        pk4d[jt][jz],
                        kernel_large=self.kernel,
                        inheritance=self.inheritance,
                        funcs=self.hfuncs,
                    )
        for jt in range(nt):
            if (self.vkernel == "linear") and (bottom_scheme == "no flux"):
                # whereever the bottom layer is masked,
                # replace it with a ghost point above it
                secondlayermasked = np.isnan(weight[:, :, 0, jt]).any(axis=1)
                # setting the value at this level zero
                weight[secondlayermasked, :, 0, jt] = 0
                shouldbemasked = np.logical_and(secondlayermasked, rz < 1 / 2)
                weight[shouldbemasked, :, 1, jt] = 0
                # setting the vertical weight of the above value to 1
                zweight[1][secondlayermasked] = 1
            for jz in range(nz):
                weight[:, :, jz, jt] *= zweight[jz]
            weight[:, :, :, jt] *= tweight[jt]

        return weight



import numpy as np
import pytest
import xarray as xr

import seaduck as sd
import seaduck.get_masks as gm
from seaduck import utils


# TODO: have a dataset that actually has maskC and is also not ECCO in the test datasets
@pytest.fixture
def masks():
    ds = utils.get_dataset("ecco")
    return gm.get_mask_arrays(sd.OceData(ds))


def test_maskC_contains_others(masks):
    # if one of the u,v,w is not defined, this cell must be a solid
    # or in other words, the maskC land is contained in other lands.
    maskC, maskU, maskV, maskW = masks
    assert np.logical_or(np.logical_not(maskC), maskU).all()
    assert np.logical_or(np.logical_not(maskC), maskV).all()
    assert np.logical_or(np.logical_not(maskC), maskW).all()


def test_not_the_same(masks):
    maskC, maskU, maskV, maskW = masks
    assert not np.allclose(maskC, maskU)
    assert not np.allclose(maskC, maskV)
    assert not np.allclose(maskC, maskW)


@pytest.mark.parametrize("od", ["ecco"], indirect=True)
def test_repeated_get_mask_array(od):
    a_mask = gm.get_masked(od, (1, 1, 1), cuvwg="U")
    assert a_mask[0] == 1
    od._ds = od._ds.drop_vars(["maskU"])
    _ = gm.get_mask_arrays(od)
    assert "maskV" in od._ds.keys()
    assert isinstance(od["maskU"], xr.DataArray)


@pytest.mark.parametrize("od", ["rect", "curv"], indirect=True)
def test_without_maskC(od):
    with pytest.warns(UserWarning):
        _, maskU, *_ = gm.get_mask_arrays(od)
    assert maskU.all()


@pytest.mark.parametrize("od", ["rect", "curv"], indirect=True)
def test_get_masked_without_maskC(od):
    with pytest.warns(UserWarning):
        hello = gm.get_masked(od, (1, 1))
    assert hello == 1
